__author__ = "Simon Nilsson"

import ast
import os
from itertools import product
from typing import Union

import pandas as pd
from sklearn.ensemble import RandomForestClassifier
from sklearn.model_selection import train_test_split

from simba.mixins.config_reader import ConfigReader
from simba.mixins.train_model_mixin import TrainModelMixin
from simba.utils.checks import (check_if_filepath_list_is_empty,
                                check_if_list_contains_values)
from simba.utils.enums import (ConfigKey, Dtypes, Formats, Methods,
                               MLParamKeys, Options, TagNames)
from simba.utils.errors import NoDataError, SamplingError
from simba.utils.printing import log_event, stdout_success
from simba.utils.read_write import (read_config_entry, read_simba_meta_files,
                                    write_df)


class GridSearchMulticlassRandomForestClassifier(ConfigReader, TrainModelMixin):

    def __init__(self, config_path: Union[str, os.PathLike]):

        ConfigReader.__init__(self, config_path=config_path)
        log_event(logger_name=str(self.__class__.__name__), log_type=TagNames.CLASS_INIT.value, msg=self.create_log_msg_from_init_args(locals=locals()))
        TrainModelMixin.__init__(self)
        self.model_dir_out = os.path.join(read_config_entry(self.config, ConfigKey.SML_SETTINGS.value, ConfigKey.MODEL_DIR.value, data_type=Dtypes.STR.value, ), "validations")
        self.bp_config = read_config_entry(config=self.config, section=ConfigKey.CREATE_ENSEMBLE_SETTINGS.value, option=ConfigKey.POSE_SETTING.value, default_value='user_defined', data_type=Dtypes.STR.value)
        if not os.path.exists(self.model_dir_out):
            os.makedirs(self.model_dir_out)
        check_if_filepath_list_is_empty(filepaths=self.target_file_paths, error_msg="Zero annotation files found in project_folder/csv/targets_inserted, cannot create models.")
        if not os.path.exists(self.configs_meta_dir):
            os.makedirs(self.configs_meta_dir)
        self.meta_file_lst = sorted(read_simba_meta_files(folder_path=self.configs_meta_dir, raise_error=True))
        print(f"Reading in {len(self.target_file_paths)} annotated files...")
        self.data_df, self.frm_idx = self.read_all_files_in_folder_mp_futures(self.target_file_paths, self.file_type)
        self.frm_idx = pd.DataFrame({"VIDEO": list(self.data_df.index), "FRAME_IDX": self.frm_idx})
        self.data_df = self.check_raw_dataset_integrity(self.data_df, logs_path=self.logs_path)
        self.data_df = self.drop_bp_cords(df=self.data_df)

    def _check_presence_of_classes_post_sampling(self, meta_dict: dict):
        for set, clf_code in product(
            [self.y_train, self.y_test], meta_dict["classifier_map"].keys()
        ):
            obs_cnt = len(self.y_df[self.y_df == clf_code])
            if obs_cnt == 0:
                raise SamplingError(
                    msg=f'Zero observations of {meta_dict["classifier_map"][clf_code]} found in the training and/or test set.'
                )

    def perform_sampling(self, meta_dict: dict):
        if (
            meta_dict[MLParamKeys.TRAIN_TEST_SPLIT_TYPE.value]
            == Methods.SPLIT_TYPE_FRAMES.value
        ):
            self.x_train, self.x_test, self.y_train, self.y_test = train_test_split(
                self.x_df, self.y_df, test_size=meta_dict["train_test_size"]
            )
        if (
            meta_dict[MLParamKeys.UNDERSAMPLE_SETTING.value]
            == "random undersample multiclass frames"
        ):
            meta_dict[MLParamKeys.UNDERSAMPLE_RATIO.value] = ast.literal_eval(
                meta_dict[MLParamKeys.UNDERSAMPLE_RATIO.value]
            )
            self.x_train, self.y_train = self.random_multiclass_frm_sampler(
                x_df=self.x_train,
                y_df=self.y_train,
                target_field=self.clf_name,
                target_var=meta_dict[MLParamKeys.UNDERSAMPLE_RATIO.value]["target_var"],
                sampling_ratio=meta_dict[MLParamKeys.UNDERSAMPLE_RATIO.value][
                    "sampling_ratio"
                ],
            )
        elif (
            meta_dict[MLParamKeys.UNDERSAMPLE_SETTING.value]
            == "random undersample multiclass bouts"
        ):
            meta_dict[MLParamKeys.UNDERSAMPLE_RATIO.value] = ast.literal_eval(
                meta_dict[MLParamKeys.UNDERSAMPLE_RATIO.value]
            )
            self.x_train, self.y_train = self.random_multiclass_bout_sampler(
                x_df=self.x_train,
                y_df=self.y_train,
                target_field=self.clf_name,
                target_var=meta_dict[MLParamKeys.UNDERSAMPLE_RATIO.value]["target_var"],
                sampling_ratio=meta_dict[MLParamKeys.UNDERSAMPLE_RATIO.value][
                    "sampling_ratio"
                ],
            )

        self._check_presence_of_classes_post_sampling(meta_dict=meta_dict)
        if meta_dict[MLParamKeys.SAVE_TRAIN_TEST_FRM_IDX.value]:
            train_data = self.frm_idx[
                self.frm_idx.index.isin(self.x_train.index)
            ].set_index("VIDEO")
            test_data = self.frm_idx[
                self.frm_idx.index.isin(self.x_test.index)
            ].set_index("VIDEO")
            write_df(
                df=train_data,
                file_type=Formats.CSV.value,
                save_path=os.path.join(
                    self.model_dir_out, f"train_idx_model_{self.config_cnt}.csv"
                ),
            )
            write_df(
                df=test_data,
                file_type=Formats.CSV.value,
                save_path=os.path.join(
                    self.model_dir_out, f"test_idx_{self.config_cnt}.csv"
                ),
            )
            print(
                f"Frame index for train and test set in model {self.config_cnt} saved in {self.model_dir_out} directory..."
            )

    def run(self):
        self.meta_dicts = self.check_validity_of_meta_files(
            data_df=self.data_df, meta_file_paths=self.meta_file_lst
        )
        if len(self.meta_dicts.keys()) == 0:
            raise NoDataError(
                msg="No valid hyper-parameter config files",
                source=self.__class__.__name__,
            )
        for config_cnt, meta_dict in self.meta_dicts.items():
            self.config_cnt = config_cnt
            self.clf_name = meta_dict[MLParamKeys.CLASSIFIER_NAME.value]
            print(
                f"Training model {config_cnt+1}/{len(self.meta_dicts.keys())} ({meta_dict[MLParamKeys.CLASSIFIER_NAME.value]})..."
            )
            self.class_names = list(meta_dict["classifier_map"].values())
            annotation_cols_to_remove = self.read_in_all_model_names_to_remove(
                self.config, self.clf_cnt, meta_dict[MLParamKeys.CLASSIFIER_NAME.value]
            )
            self.x_y_df = self.delete_other_annotation_columns(
                self.data_df, annotation_cols_to_remove, raise_error=False
            )
            self.x_df, self.y_df = self.split_df_to_x_y(
                self.x_y_df, meta_dict[MLParamKeys.CLASSIFIER_NAME.value]
            )
            self.feature_names = self.x_df.columns
            self.check_sampled_dataset_integrity(x_df=self.x_df, y_df=self.y_df)
            print(f"Number of features in dataset: {len(self.feature_names)}")
            for k, v in meta_dict["classifier_map"].items():
                print(
                    f"Number of {v} frames in dataset: {len(self.y_df[self.y_df == k])} ({(len(self.y_df[self.y_df == k]) / len(self.x_df)) * 100}%)"
                )

            check_if_list_contains_values(
                data=list(self.y_df),
                values=list(meta_dict["classifier_map"].keys()),
                name=self.clf_name,
            )
            self.perform_sampling(meta_dict=meta_dict)
            print(f"MODEL {config_cnt+1} settings")
            self.print_machine_model_information(meta_dict)
            self.rf_clf = RandomForestClassifier(
                n_estimators=meta_dict[MLParamKeys.RF_ESTIMATORS.value],
                max_features=meta_dict[MLParamKeys.RF_MAX_FEATURES.value],
                n_jobs=-1,
                criterion=meta_dict[MLParamKeys.RF_CRITERION.value],
                min_samples_leaf=meta_dict[MLParamKeys.MIN_LEAF.value],
                bootstrap=True,
                verbose=1,
                class_weight=meta_dict[MLParamKeys.CLASS_WEIGHTS.value],
            )
            print(f"Fitting {self.clf_name} model...")
            self.rf_clf = self.clf_fit(
                clf=self.rf_clf, x_df=self.x_train, y_df=self.y_train
            )
            if (
                meta_dict[MLParamKeys.PERMUTATION_IMPORTANCE.value]
                in Options.PERFORM_FLAGS.value
            ):
                self.calc_permutation_importance(x_test=self.x_test,
                                                 y_test=self.y_test,
                                                 clf=self.rf_clf,
                                                 feature_names=self.feature_names,
                                                 clf_name=self.clf_name,
                                                 save_dir=self.model_dir_out,
                                                 save_file_no=config_cnt,
                                                 plot=True)
            if (
                meta_dict[MLParamKeys.LEARNING_CURVE.value]
                in Options.PERFORM_FLAGS.value
            ):
                self.calc_learning_curve(
                    self.x_y_df,
                    self.clf_name,
                    meta_dict[MLParamKeys.LEARNING_CURVE_K_SPLITS.value],
                    meta_dict[MLParamKeys.LEARNING_CURVE_DATA_SPLITS.value],
                    meta_dict[MLParamKeys.TT_SIZE.value],
                    self.rf_clf,
                    self.model_dir_out,
                    save_file_no=config_cnt,
                    multiclass=True,
                )
            if (
                meta_dict[MLParamKeys.PRECISION_RECALL.value]
                in Options.PERFORM_FLAGS.value
            ):
                self.calc_pr_curve(
                    self.rf_clf,
                    self.x_test,
                    self.y_test,
                    self.clf_name,
                    self.model_dir_out,
                    save_file_no=config_cnt,
                )
            if (
                meta_dict[MLParamKeys.EX_DECISION_TREE.value]
                in Options.PERFORM_FLAGS.value
            ):
                self.create_example_dt(
                    self.rf_clf,
                    self.clf_name,
                    self.feature_names,
                    self.class_names,
                    self.model_dir_out,
                    save_file_no=config_cnt,
                )
            if meta_dict[MLParamKeys.CLF_REPORT.value] in Options.PERFORM_FLAGS.value:
                self.create_clf_report(
                    self.rf_clf,
                    self.x_test,
                    self.y_test,
                    self.class_names,
                    self.model_dir_out,
                    save_file_no=config_cnt,
                )
            if (
                meta_dict[MLParamKeys.IMPORTANCE_LOG.value]
                in Options.PERFORM_FLAGS.value
            ):
                self.create_x_importance_log(
                    self.rf_clf,
                    self.feature_names,
                    self.clf_name,
                    self.model_dir_out,
                    save_file_no=config_cnt,
                )
            if (
                meta_dict[MLParamKeys.IMPORTANCE_BAR_CHART.value]
                in Options.PERFORM_FLAGS.value
            ):
                self.create_x_importance_bar_chart(
                    self.rf_clf,
                    self.feature_names,
                    self.clf_name,
                    self.model_dir_out,
                    meta_dict[MLParamKeys.N_FEATURE_IMPORTANCE_BARS.value],
                    save_file_no=config_cnt)

            if MLParamKeys.SHAP_SCORES.value in meta_dict.keys():
                shap_plot = self.bp_config in {'14', '16'}
                save_n = (meta_dict[MLParamKeys.SHAP_PRESENT.value] + meta_dict[MLParamKeys.SHAP_ABSENT.value])
                shap_multiprocess = False
                if MLParamKeys.SHAP_SAVE_ITERATION.value in meta_dict.keys():
                    try:
                        save_n = int(meta_dict[MLParamKeys.SHAP_SAVE_ITERATION.value])
                    except ValueError:
                        save_n = (meta_dict[MLParamKeys.SHAP_PRESENT.value] + meta_dict[MLParamKeys.SHAP_ABSENT.value])
                if MLParamKeys.SHAP_MULTIPROCESS.value in meta_dict.keys():
                    shap_multiprocess = meta_dict[MLParamKeys.SHAP_MULTIPROCESS.value]

                if (meta_dict[MLParamKeys.SHAP_SCORES.value] in Options.PERFORM_FLAGS.value):
                    if not shap_multiprocess in Options.PERFORM_FLAGS.value:
                        self.create_shap_log(rf_clf=self.rf_clf,
                                             x=self.x_train,
                                             y=self.y_train,
                                             x_names=list(self.feature_names),
                                             clf_name=self.clf_name,
                                             cnt_present=meta_dict[MLParamKeys.SHAP_PRESENT.value],
                                             cnt_absent=meta_dict[MLParamKeys.SHAP_ABSENT.value],
                                             verbose=True,
                                             plot=shap_plot,
                                             save_it=save_n,
                                             save_dir=self.model_dir_out)
                    else:
                        self.create_shap_log_mp(rf_clf=self.rf_clf,
                                                x=self.x_train,
                                                y=self.y_train,
                                                x_names=list(self.feature_names),
                                                clf_name=self.clf_name,
                                                cnt_present=meta_dict[MLParamKeys.SHAP_PRESENT.value],
                                                cnt_absent=meta_dict[MLParamKeys.SHAP_ABSENT.value],
                                                save_dir=self.model_dir_out,
                                                save_file_suffix=config_cnt,
                                                plot=shap_plot)

                if MLParamKeys.PARTIAL_DEPENDENCY.value in meta_dict.keys():
                    if (
                        meta_dict[MLParamKeys.PARTIAL_DEPENDENCY.value]
                        in Options.PERFORM_FLAGS.value
                    ):
                        self.partial_dependence_calculator(
                            clf=self.rf_clf,
                            x_df=self.x_train,
                            clf_name=self.clf_name,
                            save_dir=self.model_dir_out,
                        )

                self.create_meta_data_csv_training_multiple_models(
                    meta_dict,
                    self.clf_name,
                    self.model_dir_out,
                    save_file_no=config_cnt,
                )
                self.save_rf_model(
                    self.rf_clf,
                    self.clf_name,
                    self.model_dir_out,
                    save_file_no=config_cnt,
                )
                print(
                    f"Classifier {self.clf_name}_{config_cnt} saved in models/validations/model_files directory ..."
                )
            stdout_success(
                msg="All models and evaluations complete. The models and evaluation files are in models/validations folders",
                source=self.__class__.__name__,
            )


# model_trainer = GridSearchMulticlassRandomForestClassifier(config_path='/Users/simon/Desktop/envs/troubleshooting/multilabel/project_folder/project_config.ini')
# model_trainer.run()



from typing import List, Optional, Tuple, Union

try:
    from typing import Literal
except:
    from typing_extensions import Literal

import functools
import multiprocessing
import os

import cv2

from simba.utils.checks import (check_if_dir_exists, check_int, check_str,
                                check_valid_boolean)
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import (find_core_cnt, get_video_meta_data,
                                    read_frm_of_video)

JPEG, PNG, WEBP = 'jpeg', 'png', 'webp'

def _video_to_frms_helper(img_batch: Tuple[int, List[int]],
                          verbose: bool,
                          img_format: str,
                          quality: int,
                          include_video_name_in_filename: bool,
                          greyscale: bool,
                          black_and_white: bool,
                          clahe: bool,
                          video_path: str,
                          save_dir: str):

    batch_cnt, frm_idxs = img_batch
    video_meta_data = get_video_meta_data(video_path=video_path)
    cap = cv2.VideoCapture(video_path)
    for frm_idx in frm_idxs:
        if include_video_name_in_filename:
            save_path = os.path.join(save_dir, f'{video_meta_data["video_name"]}_{frm_idx}.{img_format}')
        else:
            save_path = os.path.join(save_dir, f'{frm_idx}.{img_format}')
        if verbose:
            print(f"Saving image {save_path} ...")
        img = read_frm_of_video(video_path=cap, frame_index=frm_idx, greyscale=greyscale, clahe=clahe, black_and_white=black_and_white)
        if img_format == WEBP:
            cv2.imwrite(save_path, img, [cv2.IMWRITE_WEBP_QUALITY, quality])
        elif img_format == JPEG:
            cv2.imwrite(save_path, img, [cv2.IMWRITE_JPEG_QUALITY, quality])
        else:
            cv2.imwrite(save_path, img, [cv2.IMWRITE_PNG_COMPRESSION, 3])
    return batch_cnt

def video_to_frames(video_path: Union[str, os.PathLike],
                    save_dir: Union[str, os.PathLike],
                    quality: Optional[int] = 95,
                    img_format: Literal['png', 'webp'] = 'png',
                    verbose: bool = True,
                    clahe: bool = False,
                    greyscale: bool = False,
                    core_cnt: Optional[int] = -1,
                    black_and_white: bool = False,
                    include_video_name_in_filename: bool = True):


    """
    Extract all frames from a video file and save them as individual image files.

    .. note::
       Uses multiprocessing for faster frame extraction. Frames are saved with sequential numbering (0, 1, 2, ...).

    :param Union[str, os.PathLike] video_path: Path to the video file to extract frames from.
    :param Union[str, os.PathLike] save_dir: Directory where extracted frames will be saved.
    :param Optional[int] quality: Image quality for JPEG format (1-100). Higher values = better quality + larger file size. Default: 95.
    :param Literal['png', 'webp'] img_format: Format of the output images. One of: 'png', 'webp'. Default: 'png'.
    :param bool verbose: If True, prints progress messages during extraction. Default: True.
    :param bool clahe: If True, applies Contrast Limited Adaptive Histogram Equalization to each frame. Default: False.
    :param bool greyscale: If True, converts frames to greyscale. Default: False.
    :param Optional[int] core_cnt: Number of CPU cores to use for multiprocessing. -1 uses all available cores. Default: -1.
    :param bool black_and_white: If True, converts frames to black and white. Default: False.
    :param bool include_video_name_in_filename: If True, includes video name in frame filenames. Default: True.
    :return: None. Frames are saved to disk in the specified directory.

    :example:
    >>> video_to_frames(video_path=r"C:\troubleshooting\SDS_pre_post\project_folder\videos\SDI100 x ALR2 post_d7.mp4", 
    ...                save_dir=r'C:\troubleshooting\SDS_pre_post\project_folder\videos\test', 
    ...                black_and_white=False, 
    ...                verbose=True, 
    ...                img_format='webp', 
    ...                clahe=True)
    """

    timer = SimbaTimer(start=True)
    video_meta_data = get_video_meta_data(video_path=video_path)
    check_if_dir_exists(in_dir=save_dir, source=video_to_frames.__name__, raise_error=True)
    check_valid_boolean(value=verbose, source=f'{video_to_frames.__name__} verbose')
    check_valid_boolean(value=clahe, source=f'{video_to_frames.__name__} clahe')
    check_valid_boolean(value=greyscale, source=f'{video_to_frames.__name__} greyscale')
    check_valid_boolean(value=black_and_white, source=f'{video_to_frames.__name__} black_and_white')
    check_valid_boolean(value=include_video_name_in_filename, source=f'{video_to_frames.__name__} include_video_name_in_filename')
    check_int(name=f'{video_to_frames.__name__} core_cnt', value=core_cnt, min_value=-1, unaccepted_vals=[0], raise_error=True)
    check_int(name=f'{video_to_frames.__name__} quality', value=quality, min_value=1, max_value=100, raise_error=True)
    core_cnt = find_core_cnt()[0] if core_cnt -1 or core_cnt > find_core_cnt()[0] else core_cnt
    check_str(name=f'{video_to_frames.__name__} img_format', value=img_format, options=('jpeg', 'png', 'webp'))
    frm_ids = list(range(0, video_meta_data['frame_count']))
    frm_ids = [frm_ids[i * len(frm_ids) // core_cnt: (i + 1) * len(frm_ids) // core_cnt] for i in range(core_cnt)]
    frm_ids = [(i, j) for i, j in enumerate(frm_ids)]
    with multiprocessing.Pool(core_cnt, maxtasksperchild=100) as pool:
        constants = functools.partial(_video_to_frms_helper,
                                          verbose=verbose,
                                          img_format=img_format,
                                          quality=quality,
                                          greyscale=greyscale,
                                          black_and_white=black_and_white,
                                          include_video_name_in_filename=include_video_name_in_filename,
                                          video_path=video_path,
                                          clahe=clahe,
                                          save_dir=save_dir)
        for cnt, batch_id in enumerate(pool.imap(constants, frm_ids, chunksize=1)):
            if verbose:
                print(f'Video frame batch {batch_id} (of {core_cnt}) complete...')
    pool.join()
    pool.terminate()
    timer.stop_timer()
    if verbose:
        stdout_success(msg=f'All frames for video {video_path} saved in {save_dir}', elapsed_time=timer.elapsed_time_str)

# if __name__ == "__main__":
#     video_to_frames(video_path=r"C:\troubleshooting\SDS_pre_post\project_folder\videos\SDI100 x ALR2 post_d7.mp4",
#                     save_dir=r'C:\troubleshooting\SDS_pre_post\project_folder\videos\test',
#                     black_and_white=False,
#                     verbose=True,
#                     img_format='webp',
#                     clahe=True)


import io
from typing import Optional, Tuple, Union

import numpy as np
import pandas as pd
import PIL
import plotly.express as px

from simba.utils.checks import (check_float, check_instance, check_int,
                                check_valid_dataframe)
from simba.utils.data import create_color_palettes, detect_bouts
from simba.utils.errors import InvalidInputError
from simba.utils.read_write import read_df, seconds_to_timestamp


def gantt_plotly(bouts_df: pd.DataFrame,
                 img_size: Optional[Tuple[int, int]] = (640, 480),
                 bg_clr: Optional[str] = 'white',
                 title: Optional[str] = None,
                 font_size: Optional[int] = 12,
                 y_lbl: Optional[str] = 'Event',
                 x_lbl: Optional[str] = 'Session time (HH:MM:SS)',
                 show_grid: Optional[bool] = True,
                 color_palette: Optional[str] = 'Set1',
                 x_length: Optional[int] = None,
                 bar_height: Optional[float] = 0.4,
                 x_tick_spacing: Optional[Union[int, float]] = 30,
                 marker_line_color: Optional[str] = 'black',
                 marker_line_width: Optional[float] = 0.1,
                 time_format: Optional[str] = 'HH:MM:SS',
                 tick_angle: Optional[int] = 45,
                 font: Optional[str] = 'Georgia') -> np.ndarray:

    """
    Generates a Gantt chart using Plotly to visualize bout events over time.

    Creates a horizontal bar chart where each row represents an event (e.g., animal behavior) over time, with estensive customization options.

    .. image:: _static/img/gantt_plotly.webp
       :width: 600
       :align: center

    :param pd.DataFrame bouts_df: A pandas DataFrame containing the bout events data. Can be created by :func:`~simba.utils.data.detect_bouts`.
    :param Optional[Tuple[int, int]] img_size: The size of the output image as (width, height). Default (640, 480).
    :param Optional[str] bg_clr: The color of the background as a string. Deafult: white.
    :param Optional[str] title: The title of image. Deafult: None.
    :param Optional[int] font_size: The font size for the title, labels, ticks. deafult 12.
    :param Optional[str] y_lbl: The label on the y-axis. Deafult: `Event`.
    :param Optional[str] x_lbl: The label on the x-axis. Deafult: 'Session time (s)'.
    :param Optional[bool] show_grid: Whether to show the grid on the plot. Default is True.
    :param Optional[str] color_palette: The name of the color palette to use for event bars. Default: ``Set``.
    :param Optional[int] x_length: The maximum value of the x-axis. If None, the x-axis is determined based on the data.
    :param Optional[str] bar_height: The height of each bar in the Gantt chart. Default is 0.4.
    :param Optional[Union[int, float]] x_tick_spacing: The spacing between x-axis ticks. Can be either an integer or float. If float, it is treated as a fraction of the x-axis range (between 0 and 1). If integer, then interpreted as seconds. Default is 30.
    :param Optional[str] marker_line_color: Color of the bar borders. Default is 'black'.
    :param Optional[float] marker_line_width: Width of the bar borders. Default is 0.5.
    :param Optional[str] time_format: Format for x-axis tick labels. Supported formats are: - 'HH:MM:SS': Converts seconds to hours:minutes:seconds format. - 'seconds': Displays tick labels as raw seconds. Default is 'HH:MM:SS'.
    :param Optional[int] tick_angle: Angle for rotating x-axis tick labels. Default is 45 degrees.
    :param Optional[str] font: Font name. E.g., "Arial", "Verdana", "Helvetica", "Tahoma", "Trebuchet MS", "Times New Roman", "Georgia", "Courier New", "Lucida Console". Default is None, which uses the default Plotly font.
    :return: A Gantt chart image as a NumPy array (dtype=np.uint8).
    :rtype: np.ndarray

    :example:
    >>> FILE_PATH = r"D:\troubleshooting\mitra\project_folder\logs\all\592_MA147_CNO1_0515.csv"
    >>> data = read_df(file_path=FILE_PATH, file_type='csv')
    >>> bouts_df = detect_bouts(data_df=data, target_lst=list(data.columns), fps=30)
    >>> img = gantt_plotly(bouts_df=bouts_df, x_tick_spacing=120, color_palette='Set1', img_size=(1200, 700), font_size=32, show_grid=True)
    """

    check_valid_dataframe(df=bouts_df, source=f'{gantt_plotly.__name__} bouts_df', required_fields=['Start_time', 'Bout_time', 'Event'])
    check_instance(source=f'{gantt_plotly.__name__} x_tick_spacing', instance=x_tick_spacing, accepted_types=(int, float,))
    if isinstance(x_tick_spacing, float):
        check_float(name=f'{gantt_plotly.__name__} x_tick_spacing', value=x_tick_spacing, min_value=10e-6, max_value=1.0, raise_error=True)
        if x_length is not None:
            x_tick_spacing = int(x_length * x_tick_spacing)
        else:
            x_tick_spacing = int(bouts_df['End Time'].max()  * x_tick_spacing)
    elif isinstance(x_tick_spacing, int):
        check_int(name=f'{gantt_plotly.__name__} x_tick_spacing', value=x_tick_spacing, min_value=1, raise_error=True)
    check_int(name=f'{gantt_plotly.__name__} tick_angle', value=tick_angle, min_value=0, max_value=360)
    last_bout_time = bouts_df['Start_time'].max() + bouts_df['Bout_time'].max()
    if x_length is not None:
        last_bout_time = max(x_length, last_bout_time)
    tickvals = np.arange(0, last_bout_time, x_tick_spacing)
    if time_format == 'seconds':
        ticktext = [str(x) for x in tickvals]
    elif time_format == 'HH:MM:SS':
        ticktext = []
        for val in tickvals: ticktext.append(seconds_to_timestamp(val))
    else:
        raise InvalidInputError(msg=f'{time_format} is not a valid time_format', source=gantt_plotly.__name__)
    width, height = img_size
    unique_event_cnt = len(list(bouts_df['Event'].unique()))
    clrs = create_color_palettes(no_animals=1, map_size=unique_event_cnt, cmaps=[color_palette])[0]
    clrs = dict(zip(bouts_df["Event"].unique(), clrs))
    clrs = {k: f'rgb{tuple(v)}' for k, v in clrs.items()}
    fig = px.bar(bouts_df, base="Start_time", x="Bout_time", y="Event", color="Event", color_discrete_map=clrs, orientation="h")
    fig.update_layout(width=width, height=height, title=title, yaxis_type="category", showlegend=False)
    fig.update_traces(width=bar_height, marker_line_color=marker_line_color, marker_line_width=marker_line_width)

    fig.update_layout(width=width,
                      height=height,
                      title=title,
                      font=dict(
                          family=font,
                      ),
                      xaxis=dict(
                          title=dict(
                              text=x_lbl,
                              font=dict(size=font_size, family=font),
                          ),
                          showline=True,
                          linecolor="black",
                          tickvals=tickvals,
                          ticktext=ticktext,
                          tickangle=tick_angle,
                          tickfont=dict(size=font_size),
                          showgrid=show_grid,
                          range=[0, last_bout_time]),
                      yaxis=dict(
                          title=dict(
                              text=y_lbl,
                              font=dict(size=font_size)
                          ),
                          tickfont=dict(size=font_size),
                          showline=True,
                          linecolor="black",
                          showgrid=show_grid,
                      ),
                      showlegend=False)

    if bg_clr is not None:
        fig.update_layout(plot_bgcolor=bg_clr)
    img_bytes = fig.to_image(format="png")
    img = PIL.Image.open(io.BytesIO(img_bytes))
    fig = None
    return np.array(img).astype(np.uint8)


# FILE_PATH = r"D:\troubleshooting\mitra\project_folder\logs\all\592_MA147_CNO1_0515.csv"
#
# data = read_df(file_path=FILE_PATH, file_type='csv')
# bouts_df = detect_bouts(data_df=data, target_lst=list(data.columns), fps=30)
# img = gantt_plotly(bouts_df=bouts_df, x_tick_spacing=120, color_palette='Set1', img_size=(1200, 700), font_size=32, show_grid=True, bg_clr='white')
# cv2.imshow('asdasd', img)
# cv2.waitKey(1)


__author__ = "Simon Nilsson"

import os
from typing import List, Optional, Union

import numpy as np
import pandas as pd

try:
    from typing import Literal
except:
    from typing_extensions import Literal

from itertools import combinations

from simba.feature_extractors.perimeter_jit import jitted_hull
from simba.mixins.config_reader import ConfigReader
from simba.mixins.feature_extraction_mixin import FeatureExtractionMixin
from simba.mixins.feature_extraction_supplement_mixin import \
    FeatureExtractionSupplemental
from simba.mixins.train_model_mixin import TrainModelMixin
from simba.roi_tools.roi_utils import get_roi_dict_from_dfs
from simba.utils.checks import (
    check_all_file_names_are_represented_in_video_log,
    check_file_exist_and_readable, check_if_dir_exists,
    check_same_files_exist_in_all_directories, check_valid_boolean,
    check_valid_dataframe, check_valid_lst, check_video_has_rois)
from simba.utils.enums import ROI_SETTINGS, Formats
from simba.utils.errors import (DuplicationError, InvalidInputError,
                                NoFilesFoundError, NoROIDataError)
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import (copy_files_in_directory,
                                    find_files_of_filetypes_in_directory,
                                    get_fn_ext, read_df, remove_a_folder,
                                    remove_multiple_folders, write_df)

SHAPE_TYPE = "Shape_type"
TWO_POINT_BP_DISTANCES = 'TWO-POINT BODY-PART DISTANCES (MM)'
WITHIN_ANIMAL_THREE_POINT_ANGLES = 'WITHIN-ANIMAL THREE-POINT BODY-PART ANGLES (DEGREES)'
WITHIN_ANIMAL_THREE_POINT_HULL = "WITHIN-ANIMAL THREE-POINT CONVEX HULL PERIMETERS (MM)"
WITHIN_ANIMAL_FOUR_POINT_HULL = "WITHIN-ANIMAL FOUR-POINT CONVEX HULL PERIMETERS (MM)"
ANIMAL_CONVEX_HULL_PERIMETER = 'ENTIRE ANIMAL CONVEX HULL PERIMETERS (MM)'
ANIMAL_CONVEX_HULL_AREA = "ENTIRE ANIMAL CONVEX HULL AREA (MM2)"
FRAME_BP_MOVEMENT = "FRAME-BY-FRAME BODY-PART MOVEMENTS (MM)"
FRAME_BP_TO_ROI_CENTER = "FRAME-BY-FRAME BODY-PART DISTANCES TO ROI CENTERS (MM)"
FRAME_BP_INSIDE_ROI = "FRAME-BY-FRAME BODY-PARTS INSIDE ROIS (BOOLEAN)"
ARENA_EDGE = "BODY-PART DISTANCES TO VIDEO FRAME EDGE (MM)"




FEATURE_FAMILIES = [TWO_POINT_BP_DISTANCES,
                    WITHIN_ANIMAL_THREE_POINT_ANGLES,
                    WITHIN_ANIMAL_THREE_POINT_HULL,
                    WITHIN_ANIMAL_FOUR_POINT_HULL,
                    ANIMAL_CONVEX_HULL_PERIMETER,
                    ANIMAL_CONVEX_HULL_AREA,
                    FRAME_BP_MOVEMENT,
                    FRAME_BP_TO_ROI_CENTER,
                    FRAME_BP_INSIDE_ROI,
                    ARENA_EDGE]


class FeatureSubsetsCalculator(ConfigReader, TrainModelMixin):
    """
    Computes a subset of features from pose for non-ML downstream purposes.
    E.g., returns the size of animal convex hull in each frame.

    :param str config_path: path to SimBA project config file in Configparser format
    :param str save_dir: directory where to store results.
    :param List[str] feature_family: List of feature subtype to calculate. E.g., ['TWO-POINT BODY-PART DISTANCES (MM)"].
    :param bool file_checks: If true, checks that the files which the data is appended too contains the anticipated number of rows and no duplicate columns after appending. Default False.
    :param Optional[Union[str, os.PathLike]] save_dir: Directory where to save the data. If None, then the data is only appended.
    :param Optional[Union[str, os.PathLike]] data_dir: Directory of pose-estimation data to compute feature subsets for. If None, then the `/project_folder/csv/outlier_corrected_movement_locations` directory.
    :param bool append_to_features_extracted: If True, appends the data to the file sin the `features_extracted` directory. Default: False.
    :param bool append_to_targets_inserted: If True, appends the data to the file sin the `targets_inserted` directory. Default: False.

    .. note::
       `Tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/feature_subsets.md>_`

    .. image:: _static/img/feature_subsets.png
       :width: 400
       :align: center

    :example:
    >>> test = FeatureSubsetsCalculator(config_path=r"C:\troubleshooting\mitra\project_folder\project_config.ini",
    >>>                               feature_families=[FRAME_BP_MOVEMENT, WITHIN_ANIMAL_THREE_POINT_ANGLES],
    >>>                               append_to_features_extracted=False,
    >>>                               file_checks=False,
    >>>                               append_to_targets_inserted=False,
    >>>                               save_dir=r"C:\troubleshooting\mitra\project_folder\csv\new_features")
    >>> test.run()
    """

    def __init__(self,
                 config_path: Union[str, os.PathLike],
                 feature_families: List[str],
                 file_checks: bool = False,
                 save_dir: Optional[Union[str, os.PathLike]] = None,
                 data_dir: Optional[Union[str, os.PathLike]] = None,
                 append_to_features_extracted: bool = False,
                 append_to_targets_inserted: bool = False):

        check_file_exist_and_readable(file_path=config_path)
        ConfigReader.__init__(self, config_path=config_path)
        TrainModelMixin.__init__(self)
        check_valid_boolean(value=file_checks, source=f'{self.__class__.__name__} file_checks', raise_error=True)
        check_valid_boolean(value=append_to_features_extracted, source=f'{self.__class__.__name__} append_to_features_extracted', raise_error=True)
        check_valid_boolean(value=append_to_targets_inserted, source=f'{self.__class__.__name__} append_to_targets_inserted', raise_error=True)
        if save_dir is not None:
            check_if_dir_exists(in_dir=save_dir)
        check_valid_lst(data=feature_families, source=f'{self.__class__.__name__} feature_families', valid_dtypes=(str,), valid_values=FEATURE_FAMILIES, min_len=1, raise_error=True)
        self.file_checks, self.feature_families, self.save_dir = file_checks, feature_families, save_dir
        self.append_to_features_extracted = append_to_features_extracted
        self.append_to_targets_inserted = append_to_targets_inserted
        if data_dir is None:
            self.data_dir = self.outlier_corrected_dir
            self.data_paths = self.outlier_corrected_paths
        else:
            self.data_dir = data_dir
            check_if_dir_exists(in_dir=data_dir)
            self.data_paths  = find_files_of_filetypes_in_directory(directory=data_dir, extensions=['csv'], raise_error=True)
        if self.append_to_features_extracted:
            if not check_same_files_exist_in_all_directories(dirs=[self.data_dir, self.features_dir], file_type=self.file_type, raise_error=False):
                raise NoFilesFoundError(msg=f'Cannot append feature subset to files in {self.features_dir} directory: To proceed, the files in the {self.features_dir} and the {self.data_dir} directories has to contain the same number of files with the same filenames.', source=self.__class__.__name__)
        if self.append_to_targets_inserted:
            if not check_same_files_exist_in_all_directories(dirs=[self.data_dir, self.targets_folder], file_type=self.file_type, raise_error=False):
                raise NoFilesFoundError(msg=f'Cannot append feature subset to files in {self.targets_folder} directory: To proceed, the files in the {self.targets_folder} and the {self.data_dir} directories has to contain the same number of files with the same filenames.', source=self.__class__.__name__)
        self.video_names = [get_fn_ext(filepath=x)[1] for x in self.data_paths]
        for file_path in self.data_paths: check_file_exist_and_readable(file_path=file_path)
        self.temp_dir = os.path.join(self.data_dir, f"temp_data_{self.datetime}")
        if not os.path.isdir(self.temp_dir):
            os.makedirs(self.temp_dir)
        if (FRAME_BP_TO_ROI_CENTER in feature_families) or (FRAME_BP_INSIDE_ROI in feature_families):
            if not os.path.isfile(self.roi_coordinates_path):
                raise NoROIDataError(msg=f'Cannot compute ROI features: The SimBA project has no ROI data defined.')
            self.read_roi_data(); check_video_has_rois(roi_dict=self.roi_dict)
            self.roi_dict = get_roi_dict_from_dfs(rectangle_df=self.rectangles_df, circle_df=self.circles_df, polygon_df=self.polygon_df, video_name_nesting=True)
            missing_roi_videos = [x for x in self.video_names if x not in list(self.roi_dict.keys())]
            if len(missing_roi_videos) > 0:
                raise NoROIDataError(msg=f'Cannot compute ROI features: The following videos have no ROIs: {missing_roi_videos}')
        self.__get_bp_combinations()

    def __get_bp_combinations(self):
        self.two_point_combs = np.array(list(combinations(self.project_bps, 2)))
        self.within_animal_three_point_combs = {}
        self.within_animal_four_point_combs = {}
        self.animal_bps = {}
        for animal, animal_data in self.animal_bp_dict.items():
            animal_bps = [x[:-2] for x in animal_data["X_bps"]]
            self.animal_bps[animal] = animal_bps
            self.within_animal_three_point_combs[animal] = np.array(list(combinations(animal_bps, 3)))
            self.within_animal_four_point_combs[animal] = np.array(list(combinations(animal_bps, 4)))

    def _get_two_point_bp_distances(self):
        for c in self.two_point_combs:
            x1, y1, x2, y2 = list(sum([(f"{x}_x", f"{y}_y") for (x, y) in zip(c, c)], ()))
            bp1 = self.data_df[[x1, y1]].values.astype(np.float32)
            bp2 = self.data_df[[x2, y2]].values.astype(np.float32)
            self.results[f"Distance (mm) {c[0]}-{c[1]}"] = FeatureExtractionMixin.framewise_euclidean_distance(location_1=bp1, location_2=bp2, px_per_mm=self.px_per_mm)

    def __get_three_point_angles(self):
        for animal, points in self.within_animal_three_point_combs.items():
            for point in points:
                col_names = list(sum([(f"{x}_x", f"{y}_y") for (x, y) in zip(point, point)], ()))
                self.results[f"Angle (degrees) {point[0]}-{point[1]}-{point[2]}"] = (FeatureExtractionMixin.angle3pt_serialized(data=self.data_df[col_names].values))

    def __get_three_point_hulls(self):
        for animal, points in self.within_animal_three_point_combs.items():
            for point in points:
                col_names = list(sum([(f"{x}_x", f"{y}_y") for (x, y) in zip(point, point)], ()))
                three_point_arr = np.reshape(self.data_df[col_names].values, (len(self.data_df / 2), -1, 2)).astype(np.float32)
                self.results[f"{animal} three-point convex hull perimeter (mm) {point[0]}-{point[1]}-{point[2]}"] = (jitted_hull(points=three_point_arr, target=Formats.PERIMETER.value) / self.px_per_mm)

    def __get_four_point_hulls(self):
        for animal, points in self.within_animal_four_point_combs.items():
            for point in points:
                col_names = list(sum([(f"{x}_x", f"{y}_y") for (x, y) in zip(point, point)], ()))
                four_point_arr = np.reshape(self.data_df[col_names].values, (len(self.data_df / 2), -1, 2) ).astype(np.float32)
                self.results[f"{animal} four-point convex perimeter (mm) {point[0]}-{point[1]}-{point[2]}-{point[3]}"] = (jitted_hull(points=four_point_arr, target=Formats.PERIMETER.value) / self.px_per_mm)


    def __get_convex_hulls(self, method: str):
        for animal, point in self.animal_bps.items():
            col_names = list(sum([(f"{x}_x", f"{y}_y") for (x, y) in zip(point, point)], ()))
            animal_point_arr = np.reshape(self.data_df[col_names].values, (len(self.data_df / 2), -1, 2)).astype(np.float32)
            if method == 'perimeter':
                self.results[f"{animal} convex hull perimeter (mm)"] = (jitted_hull(points=animal_point_arr, target=Formats.PERIMETER.value)/ self.px_per_mm)
            else:
                self.results[f"{animal} convex hull area (mm2)"] = (jitted_hull(points=animal_point_arr, target=Formats.AREA.value) / self.px_per_mm)


    def __get_framewise_movement(self):
        for animal, animal_bps in self.animal_bps.items():
            for bp in animal_bps:
                check_valid_dataframe(df=self.data_df, source=self.file_path, required_fields=[f"{bp}_x", f"{bp}_y"])
                bp_arr = FeatureExtractionMixin.create_shifted_df(df=self.data_df[[f"{bp}_x", f"{bp}_y"]]).values
                x, y = bp_arr[:, 0:2], bp_arr[:, 2:4]
                self.results[f"{animal} movement {bp} (mm)"] = FeatureExtractionMixin.framewise_euclidean_distance(location_1=x, location_2=y, px_per_mm=self.px_per_mm)

    def __get_roi_center_distances(self):
        for animal, animal_bps in self.animal_bps.items():
            for bp in animal_bps:
                check_valid_dataframe(df=self.data_df, source=self.file_path, required_fields=[f"{bp}_x", f"{bp}_y"])
                bp_arr = self.data_df[[f"{bp}_x", f"{bp}_y"]].values.astype(np.float32)
                for roi_name, roi_data in self.roi_dict[self.video_name].items():
                    center_point = np.array([roi_data['Center_X'], roi_data['Center_Y']]).astype(np.int32)
                    distance = FeatureExtractionMixin.framewise_euclidean_distance_roi(location_1=bp_arr, location_2=center_point, px_per_mm=self.px_per_mm)
                    self.results[f"{animal} {bp} to {roi_name} center distance (mm)"] = distance

    def __get_distances_to_frm_edge(self):
        for animal, animal_bps in self.animal_bps.items():
            for bp in animal_bps:
                check_valid_dataframe(df=self.data_df, source=self.file_path, required_fields=[f"{bp}_x", f"{bp}_y"])
                bp_arr = self.data_df[[f"{bp}_x", f"{bp}_y"]].values.astype(np.float32)
                distance = FeatureExtractionSupplemental().border_distances(data=bp_arr, pixels_per_mm=self.px_per_mm, img_resolution=np.array([self.video_width, self.video_height], dtype=np.int32), time_window=1, fps=1)
                self.results[f"{animal} {bp} to left video edge distance (mm)"] = distance[:, 0]
                self.results[f"{animal} {bp} to right video edge distance (mm)"] = distance[:, 1]
                self.results[f"{animal} {bp} to top video edge distance (mm)"] = distance[:, 2]
                self.results[f"{animal} {bp} to bottom video edge distance (mm)"] = distance[:, 3]

    def __get_inside_roi(self):
        for animal, animal_bps in self.animal_bps.items():
            for bp in animal_bps:
                check_valid_dataframe(df=self.data_df, source=self.file_path, required_fields=[f"{bp}_x", f"{bp}_y"])
                bp_arr = self.data_df[[f"{bp}_x", f"{bp}_y"]].values.astype(np.float32)
                for roi_name, roi_data in self.roi_dict[self.video_name].items():
                    if roi_data[SHAPE_TYPE] == ROI_SETTINGS.RECTANGLE.value:
                        roi_coords = np.array([[roi_data['topLeftX'], roi_data['topLeftY']], [roi_data['Bottom_right_X'], roi_data['Bottom_right_Y']]])
                        r = FeatureExtractionMixin.framewise_inside_rectangle_roi(bp_location=bp_arr, roi_coords=roi_coords)
                        self.results[f"{animal} {bp} inside rectangle {roi_name} (Boolean)"] = r
                    elif roi_data[SHAPE_TYPE] == ROI_SETTINGS.CIRCLE.value:
                        circle_center = np.array([roi_data['Center_X'], roi_data['Center_Y']]).astype(np.int32)
                        r = FeatureExtractionMixin.is_inside_circle(bp=bp_arr, roi_center=circle_center, roi_radius=roi_data['radius'])
                        self.results[f"{animal} {bp} inside circle {roi_name} (Boolean)"] = r
                    elif roi_data[SHAPE_TYPE] == ROI_SETTINGS.POLYGON.value:
                        vertices = roi_data['vertices'].astype(np.int32)
                        r = FeatureExtractionMixin.framewise_inside_polygon_roi(bp_location=bp_arr, roi_coords=vertices)
                        self.results[f"{animal} {bp} inside polygon {roi_name} (Boolean)"] = r

    def __check_files(self, x: pd.DataFrame, y: pd.DataFrame, path_x: str, path_y: str):
        if len(x) != len(y):
            remove_multiple_folders(folders=[self.temp_append_dir, self.temp_dir], raise_error=False)
            raise InvalidInputError(msg=f'The files at {path_x} and {path_y} do not contain the same number of rows: {len(x)} vs {len(y)}', source=self.__class__.__name__)
        duplicated_x_cols = [i for i in x.columns if i in y.columns]
        if len(duplicated_x_cols) > 0:
            remove_multiple_folders(folders=[self.temp_append_dir, self.temp_dir], raise_error=False)
            raise DuplicationError(msg=f'Cannot append the new features to {path_y}. This file already has the following columns: {duplicated_x_cols}', source=self.__class__.__name__)

    def __append_to_data_in_dir(self, dir: Union[str, os.PathLike]):
        temp_files = find_files_of_filetypes_in_directory(directory=self.temp_dir, extensions=[f'.{self.file_type}'], as_dict=True)
        self.temp_append_dir = os.path.join(dir, f'temp_{self.datetime}')
        os.makedirs(self.temp_append_dir)
        for file_cnt, (file_name, file_path) in enumerate(temp_files.items()):
            print(f'Appending features to {file_name} ({file_cnt+1}/{len(list(temp_files.keys()))})')
            old_df = read_df(file_path=os.path.join(dir, f'{file_name}.{self.file_type}'), file_type=self.file_type).reset_index(drop=True)
            new_features_df = read_df(file_path=file_path, file_type=self.file_type).reset_index(drop=True)
            if self.file_checks:
                self.__check_files(x=new_features_df, y=old_df, path_x=file_path, path_y=os.path.join(dir, f'{file_name}.{self.file_type}'))
            save_path = os.path.join(self.temp_append_dir, f'{file_name}.{self.file_type}')
            out_df = pd.concat([old_df, new_features_df], axis=1)
            write_df(df=out_df, file_type=self.file_type, save_path=save_path)
        prior_dir = os.path.join(dir, f"Prior_to_feature_subset_append_{self.datetime}")
        os.makedirs(prior_dir)
        copy_files_in_directory(in_dir=dir, out_dir=prior_dir, filetype=self.file_type, raise_error=True)
        copy_files_in_directory(in_dir=self.temp_append_dir, out_dir=dir, filetype=self.file_type, raise_error=True)
        remove_a_folder(folder_dir=self.temp_append_dir, ignore_errors=False)

    def __append_to_targets_inserted(self, dir: Union[str, os.PathLike]):
        temp_files = find_files_of_filetypes_in_directory(directory=self.temp_dir, extensions=[f'.{self.file_type}'], as_dict=True)
        self.temp_append_dir = os.path.join(dir, f'temp_{self.datetime}')
        os.makedirs(self.temp_append_dir)
        for file_cnt, (file_name, file_path) in enumerate(temp_files.items()):
            old_df = read_df(file_path=os.path.join(dir, f'{file_name}.{self.file_type}'), file_type=self.file_type).reset_index(drop=True)
            new_features_df = read_df(file_path=file_path, file_type=self.file_type).reset_index(drop=True)
            if self.file_checks:
                self.__check_files(x=new_features_df, y=old_df, path_x=file_path, path_y=os.path.join(dir, f'{file_name}.{self.file_type}'))
            save_path = os.path.join(self.temp_append_dir, f'{file_name}.{self.file_type}')
            clf_cols = [x for x in self.clf_names if x in list(old_df.columns)]
            clf_df, old_df = old_df[clf_cols], old_df.drop(clf_cols, axis=1)
            out_df = pd.concat([old_df, new_features_df, clf_df], axis=1)
            write_df(df=out_df, file_type=self.file_type, save_path=save_path)
        prior_dir = os.path.join(dir, f"Prior_to_feature_subset_append_{self.datetime}")
        os.makedirs(prior_dir)
        copy_files_in_directory(in_dir=dir, out_dir=prior_dir, filetype=self.file_type, raise_error=True)
        copy_files_in_directory(in_dir=self.temp_append_dir, out_dir=dir, filetype=self.file_type, raise_error=True)
        remove_a_folder(folder_dir=self.temp_append_dir, ignore_errors=False)

    def run(self):
        check_all_file_names_are_represented_in_video_log(video_info_df=self.video_info_df, data_paths=self.data_paths)
        for file_cnt, file_path in enumerate(self.data_paths):
            self.file_path = file_path
            self.video_name = get_fn_ext(filepath=file_path)[1]
            video_timer = SimbaTimer(start=True)
            save_path = os.path.join(self.temp_dir, f'{self.video_name}.{self.file_type}')
            self.results = pd.DataFrame()
            print(f'Analyzing video {self.video_name}... ({file_cnt+1}/{len(self.data_paths)})')
            self.video_info, self.px_per_mm, self.fps = self.read_video_info(video_name=self.video_name)
            self.video_width, self.video_height = self.video_info['Resolution_width'].values[0], self.video_info['Resolution_height'].values[0]
            self.data_df = read_df(file_path=file_path, file_type=self.file_type)
            for family_cnt, feature_family in enumerate(self.feature_families):
                print(f"Analyzing {self.video_name} and {feature_family} (Video {file_cnt + 1}/{len(self.outlier_corrected_paths)}, Family {family_cnt + 1}/{len(self.feature_families)})...")
                if feature_family == TWO_POINT_BP_DISTANCES:
                    self._get_two_point_bp_distances()
                elif feature_family == WITHIN_ANIMAL_THREE_POINT_ANGLES:
                    self.__get_three_point_angles()
                elif feature_family == WITHIN_ANIMAL_THREE_POINT_HULL:
                    self.__get_three_point_hulls()
                elif feature_family == WITHIN_ANIMAL_FOUR_POINT_HULL:
                    self.__get_four_point_hulls()
                elif feature_family == ANIMAL_CONVEX_HULL_PERIMETER:
                    self.__get_convex_hulls(method='perimeter')
                elif feature_family == ANIMAL_CONVEX_HULL_AREA:
                    self.__get_convex_hulls(method='area')
                elif feature_family == FRAME_BP_MOVEMENT:
                    self.__get_framewise_movement()
                elif feature_family == FRAME_BP_TO_ROI_CENTER:
                    self.__get_roi_center_distances()
                elif feature_family == FRAME_BP_INSIDE_ROI:
                    self.__get_inside_roi()
                elif feature_family == ARENA_EDGE:
                    self.__get_distances_to_frm_edge()

            self.results = self.results.add_suffix('_FEATURE_SUBSET')
            self.results = self.results[sorted(self.results.columns)]
            write_df(df=self.results.fillna(-1), file_type=self.file_type, save_path=save_path)
            video_timer.stop_timer()
            print(f"Feature subsets computed for {self.video_name} complete (elapsed time {video_timer.elapsed_time_str}s)...")
        if self.append_to_features_extracted:
            print(f'Appending new feature to files in {self.features_dir}...')
            self.__append_to_data_in_dir(dir=self.features_dir)
        if self.append_to_targets_inserted:
            print(f'Appending new feature to files in {self.targets_folder}...')
            self.__append_to_targets_inserted(dir=self.targets_folder)
        if self.save_dir is not None:
            print(f"Storing new features in {self.save_dir}...")
            copy_files_in_directory(in_dir=self.temp_dir, out_dir=self.save_dir, filetype=self.file_type, raise_error=True)
        remove_a_folder(folder_dir=self.temp_dir, ignore_errors=False)
        self.timer.stop_timer()
        stdout_success(msg="Feature sub-sets calculations complete!", elapsed_time=self.timer.elapsed_time_str)




# test = FeatureSubsetsCalculator(config_path=r"C:\troubleshooting\mitra\project_folder\project_config.ini",
#                                 feature_families=[ARENA_EDGE],
#                                 append_to_features_extracted=False,
#                                 file_checks=True,
#                                 append_to_targets_inserted=False,
#                                 save_dir=r"C:\troubleshooting\mitra\project_folder\csv\feature_subset")
# test.run()

#
# test = FeatureSubsetsCalculator(config_path=r"D:\Stretch\Stretch\project_folder\project_config.ini",
#                                 feature_families=[TWO_POINT_BP_DISTANCES],
#                                 append_to_features_extracted=True,
#                                 file_checks=True,
#                                 append_to_targets_inserted=True,
#                                 save_dir=r"D:\Stretch\Stretch\project_folder\new_features")
# test.run()



# test = FeatureSubsetsCalculator(config_path=r"C:\troubleshooting\mitra\project_folder\project_config.ini",
#                                 feature_families=[TWO_POINT_BP_DISTANCES],
#                                 append_to_features_extracted=False,
#                                 file_checks=False,
#                                 append_to_targets_inserted=False,
#                                 save_dir=r"C:\troubleshooting\mitra\project_folder\csv\new_features")
# test.run()



# test = FeatureSubsetsCalculator(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                                 feature_family='Frame-by-frame body-parts inside ROIs (Boolean)',
#                                 save_dir='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/data')
# test.run()
#
#
# test = FeatureSubsetsCalculator(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                                 feature_family='Frame-by-frame body-part movements (mm)',
#                                 save_dir='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/data')
# test.run()
#
#
# test = FeatureSubsetsCalculator(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                                 feature_families=['Frame-by-frame body-part distances to ROI centers (mm)', 'Frame-by-frame body-parts inside ROIs (Boolean)'],
#                                 save_dir='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/data',
#                                 include_file_checks=True,
#                                 append_to_features_extracted=False,
#                                 append_to_targets_inserted=True)
# test.run()



__author__ = "Simon Nilsson"

import os
from pathlib import Path
from typing import Dict, Optional, Union

import cv2
import numpy as np
import pandas as pd

from simba.mixins.config_reader import ConfigReader
from simba.mixins.geometry_mixin import GeometryMixin
from simba.mixins.plotting_mixin import PlottingMixin
from simba.utils.checks import (check_instance, check_int, check_str,
                                check_that_column_exist, check_valid_boolean)
from simba.utils.data import create_color_palette
from simba.utils.enums import Formats, Options
from simba.utils.errors import CountError, InvalidFilepathError
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import (find_files_of_filetypes_in_directory,
                                    get_fn_ext, get_video_meta_data, read_df)
from simba.utils.warnings import FrameRangeWarning


class PosePlotter(object):

    """
    Create pose-estimation visualizations from data within a SimBA project folder.

    :param str data_path: Path to SimBA project directory containing pose-estimation data in parquet or CSV format.
    :param str out_directory: Directory to where to save the pose-estimation videos.
    :param int Size of the circles denoting the location of the pose-estimated body-parts.
    :param Optional[dict] clr_attr: Python dict where animals are keys and color attributes values. E.g., {'Animal_1':  (255, 107, 198)}. If None, random palettes will be used.

    """
    def __init__(self,
                 data_path: Union[str, os.PathLike],
                 out_dir: Optional[Union[str, os.PathLike]] = None,
                 palettes: Optional[Dict[str, str]] = None,
                 circle_size: Optional[int] = None,
                 bbox: Optional[bool] = False,
                 sample_time: Optional[int] = None) -> None:

        if os.path.isdir(data_path):
            config_path = os.path.join(Path(data_path).parents[1], 'project_config.ini')
        elif os.path.isfile(data_path):
            config_path = os.path.join(Path(data_path).parents[2], 'project_config.ini')
        else:
            raise InvalidFilepathError(msg=f'{data_path} not not a valid file or directory path.', source=self.__class__.__name__)
        if not os.path.isfile(config_path):
            raise InvalidFilepathError(msg=f'When visualizing pose-estimation, select an input sub-directory of the project_folder/csv folder OR file in the project_folder/csv/ANY_FOLDER directory. {data_path} does not meet these requirements and therefore SimBA cant locate the project_config.ini (expected at {config_path}', source=self.__class__.__name__)
        self.config = ConfigReader(config_path=config_path, read_video_info=False, create_logger=False)
        if os.path.isdir(data_path):
            self.files_found = find_files_of_filetypes_in_directory(directory=data_path, extensions=[f'.{self.config.file_type}'], raise_error=True)
        else:
            self.files_found = [data_path]
        self.animal_bp_dict = self.config.body_parts_lst
        if circle_size is not None:
            check_int(name='circle_size', value=circle_size, min_value=1)
        self.color_dict = {}
        check_valid_boolean(value=bbox, source=f'{self.__class__.__name__} bbox')
        if palettes is not None:
            check_instance(source=self.__class__.__name__, instance=palettes, accepted_types=(dict,))
            if len(list(palettes.keys())) != self.config.animal_cnt:
                raise CountError(msg=f'The number of color palettes ({(len(list(palettes.keys())))}) is not the same as the number of animals ({(self.config.animal_cnt)}) in the SimBA project at {self.config.project_path}')
            for c, (k, v) in enumerate(palettes.items()):
                check_str(name='palette', value=v, options=Options.PALETTE_OPTIONS_CATEGORICAL.value)
                self.color_dict[c] = create_color_palette(pallete_name=v, increments=len(self.config.body_parts_lst))
        else:
            for cnt, (k, v) in enumerate(self.config.animal_bp_dict.items()):
                self.color_dict[cnt] = self.config.animal_bp_dict[k]["colors"]
        if sample_time is not None:
            check_int(name='sample_time', value=sample_time, min_value=1)
        if out_dir is None:
            out_dir = os.path.join(os.path.dirname(self.files_found[0]), f'pose_videos_{self.config.datetime}')
        self.circle_size,self.out_dir, self.sample_time, self.bbox = (circle_size, out_dir, sample_time, bbox)
        if not os.path.exists(self.out_dir):
            os.makedirs(self.out_dir)
        self.data = {}
        for file in self.files_found:
            self.data[file] = self.config.find_video_of_file(video_dir=self.config.video_dir, filename=get_fn_ext(file)[1])

    def run(self):
        for file_cnt, (pose_path, video_path) in enumerate(self.data.items()):
            video_timer = SimbaTimer(start=True)
            video_name = get_fn_ext(pose_path)[1]
            save_video_path = os.path.join(self.out_dir, f"{video_name}.mp4")
            pose_df = read_df(file_path=pose_path, file_type=self.config.file_type, check_multiindex=True)
            video_meta_data = get_video_meta_data(video_path=video_path)
            if self.circle_size is None:
                video_circle_size = PlottingMixin().get_optimal_circle_size(frame_size=(int(video_meta_data['width']), int(video_meta_data['height'])), circle_frame_ratio=70)
            else:
                video_circle_size = self.circle_size
            if (self.sample_time is None) and (video_meta_data["frame_count"] != len(pose_df)):
                FrameRangeWarning(msg=f'The video {video_name} has pose-estimation data for {len(pose_df)} frames, but the video has {video_meta_data["frame_count"]} frames. Ensure the data and video has an equal number of frames.', source=self.__class__.__name__)
            elif isinstance(self.sample_time, int):
                sample_frm_cnt = int(video_meta_data["fps"] * self.sample_time)
                if sample_frm_cnt > len(pose_df): sample_frm_cnt = len(pose_df)
                pose_df = pose_df.iloc[0:sample_frm_cnt]
            if 'input_csv' in os.path.dirname(pose_path):
                pose_df = self.config.insert_column_headers_for_outlier_correction(data_df=pose_df, new_headers=self.config.bp_headers, filepath=pose_path)
            pose_df = (pose_df.apply(pd.to_numeric, errors="coerce").fillna(0).reset_index(drop=True))
            fourcc = cv2.VideoWriter_fourcc(*Formats.MP4_CODEC.value)
            writer = cv2.VideoWriter(save_video_path, fourcc, video_meta_data["fps"], (video_meta_data["width"], video_meta_data["height"]))
            cap = cv2.VideoCapture(video_path)
            frm_cnt = 0
            while cap.isOpened():
                ret, frame = cap.read()
                if frm_cnt >= len(pose_df):
                    break
                if ret:
                    for animal_cnt, (animal_name, animal_data) in enumerate(self.config.animal_bp_dict.items()):
                        animal_bbox = []
                        for bp_cnt, (bp_x, bp_y) in enumerate(zip(animal_data["X_bps"], animal_data["Y_bps"])):
                            check_that_column_exist(df=pose_df, column_name=[bp_x, bp_y], file_name=pose_path)
                            bp_tuple = (int(pose_df.at[frm_cnt, bp_x]), int(pose_df.at[frm_cnt, bp_y]))
                            cv2.circle(frame, bp_tuple, video_circle_size, self.color_dict[animal_cnt][bp_cnt], -1)
                            animal_bbox.append(list(bp_tuple))
                        if self.bbox and len(animal_bbox) > 4:
                            animal_bbox = GeometryMixin().keypoints_to_axis_aligned_bounding_box(keypoints=np.array(animal_bbox).reshape(-1, len(animal_bbox), 2).astype(np.int32))
                            frame = cv2.polylines(frame, [animal_bbox], True, self.color_dict[animal_cnt][0], thickness=max(1, int(video_circle_size / 1.5)), lineType=-1)
                    frm_cnt += 1
                    writer.write(frame)
                    print(f"Video: {file_cnt + 1} / {len(self.files_found)} Frame: {frm_cnt} / {len(pose_df)}")
                else:
                    print(f'Frame {frm_cnt} not found in video, terminating video creation...')
                    break
            video_timer.stop_timer()
            print(f"{save_video_path} complete... (elapsed time: {video_timer.elapsed_time_str}s)", )
            cap.release()
            writer.release()
        self.config.timer.stop_timer()
        stdout_success(msg=f"All pose videos complete. Results located in {self.out_dir} directory", elapsed_time=self.config.timer.elapsed_time_str)


# x = PosePlotter(data_path=r'C:\troubleshooting\mitra\project_folder\csv\input_csv', sample_time=100)
# x.run()
# x = PosePlotter(data_path='/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/csv/input_csv/Together_1.csv')
# x.run()


__author__ = "Simon Nilsson"

import os
from copy import deepcopy
from typing import Any, Dict, List, Optional, Union

import numpy as np
import pandas as pd

from simba.data_processors.interpolate import Interpolate
from simba.data_processors.smoothing import Smoothing
from simba.mixins.config_reader import ConfigReader
from simba.mixins.pose_importer_mixin import PoseImporterMixin
from simba.utils.checks import (check_if_dir_exists,
                                check_if_keys_exist_in_dict, check_int,
                                check_str, check_that_column_exist,
                                check_valid_lst)
from simba.utils.enums import Methods, TagNames
from simba.utils.errors import AnimalNumberError, CountError
from simba.utils.printing import SimbaTimer, log_event, stdout_success
from simba.utils.read_write import (clean_sleap_file_name,
                                    find_all_videos_in_project,
                                    find_files_of_filetypes_in_directory,
                                    get_fn_ext, get_video_meta_data, write_df)

TRACK = "track"
INSTANCE_SCORE = "instance.score"

class SLEAPImporterCSV(ConfigReader, PoseImporterMixin):
    """
    Importing SLEAP pose-estimation data into SimBA project in ``CSV`` format.

    .. note::
      `Google Colab notebook for converting SLEAP .slp to CSV written by @Toshea111  <https://colab.research.google.com/drive/1EpyTKFHVMCqcb9Lj9vjMrriyaG9SvrPO?usp=sharing>`__.
      `Example expected SLEAP csv data file for 5 animals / 4 pose-estimated body-parts  <https://github.com/sgoldenlab/simba/blob/master/misc/sleap_csv_example.csv>`__.

    :parameter str config_path: path to SimBA project config file in Configparser format
    :parameter str data_folder: Path to folder containing SLEAP data in `csv` format.
    :parameter List[str] id_lst: Animal names. This will be ignored in one animal projects and default to ``Animal_1``.
    :param Optional[Dict[str, str]] interpolation_setting: Dict defining the type and method to use to perform interpolation {'type': 'animals', 'method': 'linear'}.
    :param Optional[Dict[str, Union[str, int]]] smoothing_settings: Dictionary defining the pose estimation smoothing method {'time_window': 500, 'method': 'gaussian'}.

    References
    ----------
    .. [1] Pereira et al., SLEAP: A deep learning system for multi-animal pose tracking, `Nature Methods`, 2022.

    >>> sleap_csv_importer = SLEAPImporterCSV(config_path=r'project_folder/project_config.ini', data_folder=r'data_folder', id_lst=['Termite_1', 'Termite_2', 'Termite_3', 'Termite_4', 'Termite_5'], interpolation_settings={'type': 'animals', 'method': 'linear'}, smoothing_settings = {'time_window': 500, 'method': 'gaussian'})
    >>> sleap_csv_importer.run()
    """

    def __init__(self,
                 config_path: Union[str, os.PathLike],
                 data_folder: Union[str, os.PathLike],
                 id_lst: List[str],
                 interpolation_settings: Optional[Dict[str, str]] = None,
                 smoothing_settings: Optional[Dict[str, Union[int, str]]] = None):

        ConfigReader.__init__(self, config_path=config_path, read_video_info=False)
        PoseImporterMixin.__init__(self)
        check_if_dir_exists(in_dir=data_folder)
        check_valid_lst(data=id_lst, source=f'{self.__class__.__name__} id_lst', valid_dtypes=(str,), min_len=1)
        if interpolation_settings is not None:
            check_if_keys_exist_in_dict(data=interpolation_settings, key=['method', 'type'], name=f'{self.__class__.__name__} interpolation_settings')
            check_str(name=f'{self.__class__.__name__} interpolation_settings type', value=interpolation_settings['type'], options=('body-parts', 'animals'))
            check_str(name=f'{self.__class__.__name__} interpolation_settings method', value=interpolation_settings['method'], options=('linear', 'quadratic', 'nearest'))
        if smoothing_settings is not None:
            check_if_keys_exist_in_dict(data=smoothing_settings, key=['method', 'time_window'], name=f'{self.__class__.__name__} smoothing_settings')
            check_str(name=f'{self.__class__.__name__} smoothing_settings method', value=smoothing_settings['method'], options=('savitzky-golay', 'gaussian'))
            check_int(name=f'{self.__class__.__name__} smoothing_settings time_window', value=smoothing_settings['time_window'], min_value=1)

        log_event(logger_name=str(__class__.__name__), log_type=TagNames.CLASS_INIT.value, msg=self.create_log_msg_from_init_args(locals=locals()))
        self.interpolation_settings, self.smoothing_settings = (interpolation_settings, smoothing_settings)
        self.data_folder, self.id_lst = data_folder, id_lst
        self.import_log_path = os.path.join(self.logs_path, f"data_import_log_{self.datetime}.csv")
        self.video_paths = find_all_videos_in_project(videos_dir=self.video_dir)
        self.input_data_paths = find_files_of_filetypes_in_directory(directory=self.data_folder, extensions=['.csv'], raise_error=True)
        if self.pose_setting is Methods.USER_DEFINED.value:
            self.__update_config_animal_cnt()
        if self.animal_cnt > 1:
            self.data_and_videos_lk = self.link_video_paths_to_data_paths(data_paths=self.input_data_paths, video_paths=self.video_paths, filename_cleaning_func=clean_sleap_file_name)
            self.check_multi_animal_status()
            self.animal_bp_dict = self.create_body_part_dictionary(self.multi_animal_status, self.id_lst, self.animal_cnt, self.x_cols, self.y_cols, self.p_cols, self.clr_lst)
            if self.pose_setting is Methods.USER_DEFINED.value:
                self.update_bp_headers_file(update_bp_headers=True)
        else:
            self.data_and_videos_lk = dict([(get_fn_ext(file_path)[1], {"DATA": file_path, "VIDEO": None}) for file_path in self.input_data_paths])
        print(f"Importing {len(list(self.data_and_videos_lk.keys()))} SLEAP CSV file(s)...")

    def run(self):
        for file_cnt, (video_name, video_data) in enumerate(self.data_and_videos_lk.items()):
            output_filename = clean_sleap_file_name(filename=video_name)
            print(f"Importing {output_filename}...")
            video_timer = SimbaTimer(start=True)
            self.video_name = video_name
            self.save_path = os.path.join(os.path.join(self.input_csv_dir, f"{output_filename}.{self.file_type}"))
            data_df = pd.read_csv(video_data["DATA"])
            if INSTANCE_SCORE in data_df.columns:
                data_df = data_df.drop([INSTANCE_SCORE], axis=1)
            idx = data_df.iloc[:, :2]
            check_that_column_exist(df=idx, column_name=TRACK, file_name=video_name)
            data_unique_tracks = list(idx[TRACK].unique())
            if len(data_unique_tracks) != self.animal_cnt:
                raise AnimalNumberError(msg=f'The SLEAP CSV file {video_data["DATA"]} contains data for {len(data_unique_tracks)} tracks (found tracks: {data_unique_tracks}). The SimBA project config says the SimBA project expects data for {self.animal_cnt} animals.')
            if len(data_unique_tracks) == 1:
                idx[TRACK] = 'track_0'; data_unique_tracks = ['track_0']
            idx[TRACK] = idx[TRACK].fillna(data_unique_tracks[0])
            idx[TRACK] = idx[TRACK].astype(str).str.replace(r"[^\d.]+", "", regex=True).astype(int)
            data_df = data_df.iloc[:, 2:].fillna(0)
            if self.animal_cnt > 1:
                self.data_df = pd.DataFrame(self.transpose_multi_animal_table(data=data_df.values, idx=idx.values, animal_cnt=self.animal_cnt))
            else:
                idx = list(idx.drop(TRACK, axis=1)["frame_idx"])
                self.data_df = data_df.set_index([idx]).sort_index()
                self.data_df.columns = np.arange(len(self.data_df.columns))
                self.data_df = self.data_df.reindex(range(0, self.data_df.index[-1] + 1), fill_value=0)
            if len(self.bp_headers) != len(self.data_df.columns):
                raise CountError(msg=f"The SimBA project expects {len(self.bp_headers)} data columns, but your SLEAP data file at {video_data['DATA']} contains {len(self.data_df.columns)} columns.", source=self.__class__.__name__)
            self.data_df.columns = self.bp_headers
            self.out_df = deepcopy(self.data_df)
            if self.animal_cnt > 1:
                self.initialize_multi_animal_ui(animal_bp_dict=self.animal_bp_dict,
                                                video_info=get_video_meta_data(video_data["VIDEO"]),
                                                data_df=self.data_df,
                                                video_path=video_data["VIDEO"])
                self.multianimal_identification()
            write_df(df=self.out_df, file_type=self.file_type, save_path=self.save_path, multi_idx_header=True)
            if self.interpolation_settings is not None:
                interpolator = Interpolate(config_path=self.config_path, data_path=self.save_path, type=self.interpolation_settings['type'], method=self.interpolation_settings['method'], multi_index_df_headers=True, copy_originals=False)
                interpolator.run()
            if self.smoothing_settings is not None:
                smoother = Smoothing(config_path=self.config_path, data_path=self.save_path, time_window=self.smoothing_settings['time_window'], method=self.smoothing_settings['method'], multi_index_df_headers=True, copy_originals=False)
                smoother.run()
            video_timer.stop_timer()
            stdout_success(msg=f"Video {video_name} data imported...", elapsed_time=video_timer.elapsed_time_str, source=self.__class__.__name__)
        self.timer.stop_timer()
        stdout_success(msg=f"{len(list(self.data_and_videos_lk.keys()))} file(s) imported to the SimBA project {self.input_csv_dir}", source=self.__class__.__name__)



# test = SLEAPImporterCSV(config_path=r"D:\troubleshooting\two_animals_sleap\project_folder\project_config.ini",
#                  data_folder=r'D:\troubleshooting\two_animals_sleap\import_data',
#                  id_lst=['Animal_1', 'Animal_2'],
#                  interpolation_settings=None,
#                  smoothing_settings = None)
# test.run()


# test = SLEAPImporterCSV(config_path=r"C:\troubleshooting\SLP_TEST0525\project_folder\project_config.ini",
#                  data_folder=r'C:\troubleshooting\SLP_TEST0525\data_csv',
#                  id_lst=['Animal_1', 'Animal_2', 'Animal_3', 'Animal_4'],
#                  interpolation_settings={'type': 'body-parts', 'method': 'linear'},
#                  smoothing_settings = {'time_window': 500, 'method': 'gaussian'})
# test.run()
#



#
# test = SLEAPImporterCSV(config_path=r"C:\troubleshooting\sleap_import_two_tracks\project_folder\project_config.ini",
#                  data_folder=r'C:\troubleshooting\sleap_import_two_tracks\data_csv',
#                  id_lst=['Track_0', 'Track_1'],
#                  interpolation_settings={'type': 'animals', 'method': 'linear'},
#                  smoothing_settings = {'time_window': 500, 'method': 'gaussian'})
# test.run()
# #



# test = SLEAPImporterCSV(config_path=r"C:\troubleshooting\sleap_import_two_tracks\project_folder\project_config.ini",
#                  data_folder=r'C:\troubleshooting\sleap_import_two_tracks\data_csv',
#                  id_lst=['Track_0', 'Track_1'],
#                  interpolation_settings={'type': 'animals', 'method': 'linear'},
#                  smoothing_settings = {'time_window': 500, 'method': 'gaussian'})
# test.run()

# test = SLEAPImporterCSV(config_path=r'/Users/simon/Desktop/envs/troubleshooting/Hornet/project_folder/project_config.ini',
#                  data_folder=r'/Users/simon/Desktop/envs/troubleshooting/Hornet_single_slp/import',
#                  id_lst=['Hornet'],
#                  interpolation_settings="Body-parts: Nearest",
#                  smoothing_settings = {'Method': 'None', 'Parameters': {'Time_window': '200'}})
# test.run()

# test = SLEAPImporterCSV(config_path=r'/Users/simon/Desktop/envs/troubleshooting/Hornet/project_folder/project_config.ini',
#                  data_folder=r'/Users/simon/Desktop/envs/troubleshooting/Hornet_single_slp/import',
#                  id_lst=['Hornet'],
#                  interpolation_settings="Body-parts: Nearest",
#                  smoothing_settings = {'Method': 'None', 'Parameters': {'Time_window': '200'}})
# test.run()

# test = SLEAPImporterCSV(config_path=r'/Users/simon/Desktop/envs/troubleshooting/slp_1_animal_1_bp/project_folder',
#                  data_folder='/Users/simon/Desktop/envs/troubleshooting/slp_1_animal_1_bp/import',
#                  id_lst=['Termite_1'],
#                  interpolation_settings="Body-parts: Nearest",
#                  smoothing_settings = {'Method': 'Savitzky Golay', 'Parameters': {'Time_window': '200'}})
# test.run()



__author__ = "Simon Nilsson"

import os
from typing import List, Optional, Union

import numpy as np
import pandas as pd

from simba.mixins.config_reader import ConfigReader
from simba.mixins.feature_extraction_mixin import FeatureExtractionMixin
from simba.roi_tools.ROI_directing_analyzer import DirectingROIAnalyzer
from simba.utils.checks import (
    check_all_file_names_are_represented_in_video_log,
    check_file_exist_and_readable, check_that_column_exist,
    check_valid_boolean, check_valid_lst)
from simba.utils.data import slice_roi_dict_for_video
from simba.utils.enums import Keys, TagNames
from simba.utils.errors import (BodypartColumnNotFoundError, CountError,
                                InvalidFilepathError, InvalidInputError,
                                NoFilesFoundError, ROICoordinatesNotFoundError)
from simba.utils.printing import SimbaTimer, log_event, stdout_success
from simba.utils.read_write import (find_files_of_filetypes_in_directory,
                                    get_fn_ext, read_data_paths, read_df,
                                    write_df)
from simba.utils.warnings import DuplicateNamesWarning, ROIWarning


class ROIFeatureCreator(ConfigReader, FeatureExtractionMixin):
    """
    Compute features based on the relationships between the location of the animals and the location of
    user-defined ROIs. This includes the distance to the ROIs, if the animals are inside the ROIs, and if the
    animals are directing towards the ROIs (if viable)

    .. note::
        `ROI tutorials <https://github.com/sgoldenlab/simba/blob/master/docs/ROI_tutorial_new.md>`__.

    :param Union[str, os.PathLike] config_path: Path to SimBA project config file in Configparser format.
    :param List[str] body_parts: List of the body-parts to use as proxy for animal location(s).
    :param Optional[Union[str, os.PathLike]] data_path: Path to folder or file holding the data used to calculate ROI aggregate statistics. If None, then defaults to the `project_folder/csv/outlier_corrected_movement_location` directory of the SimBA project. Default: None.
    :param Optional[bool] append_data: If True, adds the features to the data in the `project_folder/csv/features_extracted` directory. Else, the data is held  in memory.


    :example:
    >>> roi_featurizer = ROIFeatureCreator(config_path='/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini', body_parts=['Nose_1', 'Nose_2'])
    >>> roi_featurizer.run()
    >>> roi_featurizer.save()

    >>> roi_featurizer = ROIFeatureCreator(config_path=r"C:\troubleshooting\spontenous_alternation\project_folder\project_config.ini", body_parts=['nose'], data_path=r"C:\troubleshooting\spontenous_alternation\project_folder\csv\outlier_corrected_movement_location\F1 HAB.csv", append_data=True)
    >>> roi_featurizer.run()
    >>> roi_featurizer.save()
    """

    def __init__(self,
                 config_path: Union[str, os.PathLike],
                 body_parts: List[str],
                 data_path: Optional[Union[str, os.PathLike]] = None,
                 append_data: bool = False):

        check_valid_lst(data=body_parts, source=f"{self.__class__.__name__} body-parts", valid_dtypes=(str,), min_len=1)
        if len(set(body_parts)) != len(body_parts):
            raise CountError(msg=f"All body-part entries have to be unique. Got {body_parts}", source=self.__class__.__name__)
        log_event(logger_name=str(__class__.__name__), log_type=TagNames.CLASS_INIT.value, msg=self.create_log_msg_from_init_args(locals=locals()))
        ConfigReader.__init__(self, config_path=config_path)
        FeatureExtractionMixin.__init__(self, config_path=config_path)
        check_valid_boolean(value=[append_data], source=f'{self.__class__.__name__} append_data', raise_error=True)
        if data_path is None:
            self.data_dir = self.outlier_corrected_dir
            self.data_paths = find_files_of_filetypes_in_directory(directory=self.data_dir, extensions=[f'.{self.file_type}'], raise_error=False)
        elif os.path.isdir(data_path):
            self.data_dir = data_path
            self.data_paths = find_files_of_filetypes_in_directory(directory=self.data_dir, extensions=[f'.{self.file_type}'], raise_error=False)
        elif os.path.isfile(data_path):
            check_file_exist_and_readable(file_path=data_path)
            dir, _, ext = get_fn_ext(filepath=data_path)
            if ext != f'.{self.file_type}':
                raise InvalidFilepathError(msg=f'{data_path} is not a valid {self.file_type} file', source=self.__class__.__name__)
            self.data_dir = dir
            self.data_paths = [data_path]
        else:
            raise InvalidInputError(msg=f'{data_path} is not a valid data_path', source=self.__class__.__name__)
        if len(self.data_paths) == 0:
            raise NoFilesFoundError(msg=f"No data found in the {self.data_dir} directory", source=self.__class__.__name__)
        if not os.path.isfile(self.roi_coordinates_path):
            raise ROICoordinatesNotFoundError(expected_file_path=self.roi_coordinates_path)
        self.read_roi_data()
        self.roi_directing_viable = self.check_directionality_viable()[0]
        for bp in body_parts:
            if bp not in self.body_parts_lst:
                raise BodypartColumnNotFoundError(msg=f"The body-part {bp} is not a valid body-part in the SimBA project. Options: {self.body_parts_lst}", source=self.__class__.__name__)
        self.bp_lk = {}
        for cnt, bp in enumerate(body_parts):
            animal = self.find_animal_name_from_body_part_name(bp_name=bp, bp_dict=self.animal_bp_dict)
            self.bp_lk[cnt] = [animal, bp, [f'{bp}_{"x"}', f'{bp}_{"y"}', f'{bp}_{"p"}']]
        if self.roi_directing_viable:
            print("Directionality calculations are VIABLE.")
            self.directing_analyzer = DirectingROIAnalyzer(config_path=config_path, data_path=self.data_paths)
            self.directing_analyzer.run()
            self.dr = self.directing_analyzer.results_df
        else:
            print("Directionality calculations are NOT VIABLE.")
            self.directing_analyzer = None
            self.dr = None
        self.append_data = append_data
        if append_data and len(self.feature_file_paths) == 0:
            raise NoFilesFoundError(msg=f"No data found in the {self.features_dir} directory. Create feature data before appending ROI features.", source=self.__class__.__name__)
        print(f"Processing {len(self.data_paths)} video(s) for ROI features...")


    def run(self):
        check_all_file_names_are_represented_in_video_log(video_info_df=self.video_info_df, data_paths=self.data_paths)
        self.summary = pd.DataFrame(columns=["VIDEO", "ANIMAL", "SHAPE NAME", "MEASUREMENT", "VALUE"])
        if self.append_data:
            data_filenames = set([get_fn_ext(x)[1] for x in self.data_paths])
            feature_extraction_filenames = set([get_fn_ext(x)[1] for x in self.feature_file_paths])
            missing_data_files = [x for x in data_filenames if x not in feature_extraction_filenames]
            missing_feature_files = [x for x in feature_extraction_filenames if x not in data_filenames]
            if len(missing_feature_files) > 0:
                raise NoFilesFoundError(msg=f"Before appending ROI features, make sure each video is represented in both the {self.data_dir} and {self.features_dir} directory. You have videos represented in the {self.data_dir} that does not exist in the {self.features_dir}: {missing_feature_files}", source=self.__class__.__name__)
            elif len(missing_data_files) > 0:
                raise NoFilesFoundError(msg=f"Before appending ROI features, make sure each video is represented in both the {self.data_dir} and {self.features_dir} directory. You have videos represented in the {self.features_dir} that does not exist in the {self.data_dir}: {missing_data_files}", source=self.__class__.__name__)

        for file_cnt, file_path in enumerate(self.data_paths):
            video_timer = SimbaTimer(start=True)
            _, self.video_name, _ = get_fn_ext(file_path)
            _, self.pixels_per_mm, self.fps = self.read_video_info(video_name=self.video_name)
            data_df = read_df(file_path=file_path, file_type=self.file_type)
            self.video_roi_dict, self.shape_names = slice_roi_dict_for_video(data=self.roi_dict, video_name=self.video_name)
            if len(self.shape_names) == 0:
                ROIWarning(msg=f'No ROIs detected for video {self.video_name}. Skipping ROI feature calculations for video {self.video_name}', source=self.__class__.__name__)
                continue
            else:
                self.out_df = pd.DataFrame()
                for animal_cnt, animal_data in self.bp_lk.items():
                    animal_name, body_part_name, bp_cols = animal_data
                    check_that_column_exist(df=data_df, column_name=bp_cols, file_name=file_path)
                    animal_df = data_df[bp_cols]
                    for _, row in self.video_roi_dict[Keys.ROI_RECTANGLES.value].iterrows():
                        roi_name, roi_center = row["Name"], np.array([row["Center_X"], row["Center_Y"]])
                        roi_border = np.array([[row["topLeftX"], row["topLeftY"]], [row["Bottom_right_X"], row["Bottom_right_Y"]]])
                        c = f"{roi_name} {animal_name} {body_part_name} distance"
                        self.out_df[c] = FeatureExtractionMixin.framewise_euclidean_distance_roi(location_1=animal_df.values[:, 0:2], location_2=roi_center, px_per_mm=self.pixels_per_mm)
                        self.summary.loc[len(self.summary)] = [self.video_name, animal_name, roi_name, "Average distance (mm)", round(float(self.out_df[c].mean()), 4)]
                        c = f"{roi_name} {animal_name} {body_part_name} in zone"
                        self.out_df[c] = FeatureExtractionMixin.framewise_inside_rectangle_roi(bp_location=animal_df.values[:, 0:2], roi_coords=roi_border)
                    for _, row in self.video_roi_dict[Keys.ROI_CIRCLES.value].iterrows():
                        roi_center = np.array([row["centerX"], row["centerY"]])
                        roi_name, radius = row["Name"], row["radius"]
                        c = f"{roi_name} {animal_name} {body_part_name} distance"
                        self.out_df[c] = FeatureExtractionMixin.framewise_euclidean_distance_roi(location_1=animal_df.values[:, 0:2], location_2=roi_center, px_per_mm=self.pixels_per_mm)
                        self.summary.loc[len(self.summary)] = [self.video_name, animal_name, roi_name, "Average distance (mm)", round(float(self.out_df[c].mean()), 4)]
                        in_zone_col = f"{roi_name} {animal_name} {body_part_name} in zone"
                        self.out_df[in_zone_col] = 0
                        self.out_df.loc[self.out_df[c] <= (row["radius"] / self.pixels_per_mm), in_zone_col] = 1
                    for _, row in self.video_roi_dict[Keys.ROI_POLYGONS.value].iterrows():
                        roi_vertices = np.array(list(zip(row["vertices"][:, 0], row["vertices"][:, 1])))
                        roi_name, roi_center = row["Name"], np.array([row["Center_X"], row["Center_Y"]])
                        c = f"{roi_name} {animal_name} {body_part_name} distance"
                        self.out_df[c] = FeatureExtractionMixin.framewise_euclidean_distance_roi(location_1=animal_df.values[:, 0:2], location_2=roi_center, px_per_mm=self.pixels_per_mm)
                        self.summary.loc[len(self.summary)] = [self.video_name, animal_name, roi_name, "Average distance (mm)", round(float(self.out_df[c].mean()), 4)]
                        c = f"{roi_name} {animal_name} {body_part_name} in zone"
                        self.out_df[c] = FeatureExtractionMixin.framewise_inside_polygon_roi(bp_location=animal_df.values[:, 0:2], roi_coords=roi_vertices)
                    if self.roi_directing_viable:
                        animal_dr = self.dr.loc[(self.dr["Video"] == self.video_name) & (self.dr["Animal"] == animal_name)]
                        for shape_name in self.shape_names:
                            animal_shape_idx = list(animal_dr.loc[(animal_dr["ROI"] == shape_name) & (animal_dr["Directing_BOOL"] == 1)]["Frame"])
                            c = f"{shape_name} {animal_name} facing"
                            self.out_df[c] = 0
                            self.out_df.loc[animal_shape_idx, c] = 1
                            self.summary.loc[len(self.summary)] = [self.video_name, animal_name, shape_name, "Total direction time (s)", round((float(self.out_df[c].sum()) / self.fps), 4)]
                video_timer.stop_timer()
                if self.append_data:
                    feature_path = os.path.join(self.features_dir, f'{self.video_name}.{self.file_type}')
                    features_df = read_df(file_path=feature_path, file_type=self.file_type)
                    duplicated_columns = [x for x in features_df.columns if x in self.out_df.columns]
                    if len(duplicated_columns) > 0:
                        DuplicateNamesWarning(msg=f'Some new ROI feature column names already exist in the {feature_path} file and have been duplicated: {duplicated_columns}', source=self.__class__.__name__)
                    self.out_df = pd.concat([features_df, self.out_df], axis=1).reset_index(drop=True)
                    write_df(df=self.out_df, file_type=self.file_type, save_path=feature_path)
                    print(f"New file with ROI features created at  {feature_path} saved (File {file_cnt+1}/{len(self.data_paths)}), elapsed time: {video_timer.elapsed_time_str}s")
        self.timer.stop_timer()
        stdout_success(msg=f"ROI features analysed for {len(self.data_paths)} videos", elapsed_time=self.timer.elapsed_time_str)

    def save(self):
        save_path = os.path.join(self.logs_path, f"ROI_features_summary_{self.datetime}.csv")
        self.summary.to_csv(save_path)
        print(f"ROI feature summary data saved at {save_path}")
        self.timer.stop_timer()
        if self.append_data:
            stdout_success(msg=f"{len(self.data_paths)} new file(s) with ROI features saved in {self.features_dir}", elapsed_time=self.timer.elapsed_time_str)
        else:
            stdout_success(msg=f"{len(self.data_paths)} data files analyzed for ROI features", elapsed_time=self.timer.elapsed_time_str)


# #roi_featurizer = ROIFeatureCreator(config_path=r"C:\troubleshooting\roi_feature_issue\project_folder\project_config.ini",
#                                    body_parts=['nose'],
#                                    data_path=r"C:\troubleshooting\roi_feature_issue\project_folder\csv\outlier_corrected_movement_location\20250130_Oxyipn_Vls_D4_Sst-107.csv",
#                                    append_data=True)
# #roi_featurizer.run()
# #roi_featurizer.save()




# roi_featurizer = ROIFeatureCreator(config_path=r"C:\troubleshooting\spontenous_alternation\project_folder\project_config.ini",
#                                    body_parts=['nose'],
#                                    data_path=r"C:\troubleshooting\spontenous_alternation\project_folder\csv\outlier_corrected_movement_location\F1 HAB.csv",
#                                    append_data=True)
# roi_featurizer.run()
# roi_featurizer.save()




#
# roi_featurizer = ROIFeatureCreator(config_path='/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                                    body_parts=['Nose_1', 'Nose_2'],
#                                    data_path='/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/csv/outlier_corrected_movement_location/Together_1.csv')
# roi_featurizer.run()


# roi_featurizer = ROIFeatureCreator(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini')
# roi_featurizer.roi_directing_viable
# roi_featurizer.run()
# roi_featurizer.save()


# roi_featurizer = ROIFeatureCreator(config_path='/Users/simon/Desktop/envs/troubleshooting/two_animals_16bp_032023/project_folder/project_config.ini')
# roi_featurizer.run()
# roi_featurizer.save()



__author__ = "Simon Nilsson"

import glob
import os
import subprocess
from typing import Optional, Tuple, Union

import cv2
import numpy as np
import pandas as pd

try:
    from typing import Literal
except:
    from typing_extensions import Literal

from simba.mixins.plotting_mixin import PlottingMixin
from simba.utils.checks import (check_if_dir_exists, check_int, check_str,
                                check_valid_boolean)
from simba.utils.enums import Formats, Options, TextOptions
from simba.utils.errors import CountError
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import (find_files_of_filetypes_in_directory,
                                    get_fn_ext, get_video_meta_data,
                                    read_frm_of_video)
from simba.video_processors.roi_selector import ROISelector
from simba.video_processors.video_processing import crop_video


class MultiCropper(object):
    """
    Crop each video of a specific file format (e.g., mp4) in a directory into N smaller cropped videos.


    .. note::
       `Multi-crop tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/Tutorial_tools.md#multi-crop-videos>`__.

       `Expected GPU timesavings <https://github.com/sgoldenlab/simba/blob/master/docs/gpu_vs_cpu_video_processing_runtimes.md>`__.

    .. image:: _static/img/multicrop.gif
       :width: 700
       :align: center

    :param Union[str, os.PathLike] input_folder: Folder path holding videos to be cropped.
    :param Literal['avi', 'mp4', 'mov', 'flv', 'm4v'] file_type: File type of input video files inside the ``input_folder`` directory.
    :param Union[str, os.PathLike] output_folder: Directory where to store the cropped videos.
    :param int crop_cnt: The number of cropped videos to create from each input video. Minimum: 2.
    :param bool gpu: If True, use GPU codecs, else CPU. Default CPU.


    :example:
    >>> cropper = MultiCropper(file_type='mp4', input_folder=r'C:\troubleshooting\mitra\test', output_folder=r'C:\troubleshooting\mitra\test\cropped', crop_cnt=2, gpu=True)
    >>> cropper.run()
    """

    def __init__(self,
                 file_type: Literal["avi", "mp4", "mov", "flv", "m4v"],
                 input_folder: Union[str, os.PathLike],
                 output_folder: Union[str, os.PathLike],
                 crop_cnt: int,
                 gpu: bool = False,
                 quality: int = 60):

        check_str(name="FILE TYPE", value=file_type.lower(), options=Options.ALL_VIDEO_FORMAT_OPTIONS_2.value)
        check_if_dir_exists(in_dir=input_folder)
        check_if_dir_exists(in_dir=output_folder)
        check_int(name=f'{self.__class__.__name__} crop_cnt', value=crop_cnt, min_value=2)
        check_int(name=f'{crop_video.__name__} quality', value=quality, min_value=1, max_value=100)
        check_valid_boolean(value=[gpu], source=f'{self.__class__.__name__} gpu')
        self.file_type, self.crop_cnt, self.gpu, self.quality = file_type.lower(), crop_cnt, gpu, quality
        self.input_folder, self.output_folder = input_folder, output_folder
        self.video_paths = find_files_of_filetypes_in_directory(directory=self.input_folder, extensions=[f'.{file_type}'], raise_error=True)
        self.font = Formats.FONT.value
        self.crop_df = pd.DataFrame(columns=["video", "height", "width", "top_left_x", "top_left_y"])

    def __test_crop_locations(self):
        for idx, row in self.crop_df.iterrows():
            lst = [row["height"], row["width"], row["top_left_x"], row["top_left_y"]]
            video_name = row["video"]
            if all(v == 0 for v in lst):
                raise CountError(msg=f"SIMBA ERROR: A crop for video {video_name} has all crop coordinates set to zero. Did you click ESC, space or enter before defining the rectangle crop coordinates?!", source=self.__class__.__name__)
            else:
                pass

    def run(self):
        for file_cnt, video_path in enumerate(self.video_paths):
            video_meta_data = get_video_meta_data(video_path=video_path)
            _, video_name, _ = get_fn_ext(filepath=video_path)
            font_size, x_space, y_space = PlottingMixin().get_optimal_font_scales(text=f"Define the ROI boundaries of {self.crop_cnt} cropped videos", accepted_px_width=video_meta_data['width'], accepted_px_height=int(video_meta_data['height']/10), text_thickness=3, font=self.font)
            img = read_frm_of_video(video_path=video_path, frame_index=0)
            original_img = np.copy(img)
            txt_pos = (10, 20)
            img = PlottingMixin().put_text(img=img, text=video_name, pos=txt_pos, font_size=font_size, font_thickness=2, font=self.font, text_color=TextOptions.COLOR.value, text_color_bg=(0, 0, 0))
            img = PlottingMixin().put_text(img=img, text=f"Define the ROI boundaries of {self.crop_cnt} cropped videos", pos=(txt_pos[0], txt_pos[1] + (y_space*1)), font_size=font_size, font_thickness=2, font=self.font, text_color=TextOptions.COLOR.value, text_color_bg=(0, 0, 0))
            img = PlottingMixin().put_text(img=img, text=f"Press ESC to continue", pos=(txt_pos[0], txt_pos[1] + (y_space*2)), font_size=font_size, font_thickness=2, font=self.font, text_color=(0, 0, 255), text_color_bg=(0, 0, 0), text_bg_alpha=1.0)
            cv2.namedWindow("VIDEO IMAGE", cv2.WINDOW_NORMAL)
            cv2.imshow("VIDEO IMAGE", img)
            while True:
                k = cv2.waitKey(1) & 0xFF
                if k == 27:
                    cv2.destroyAllWindows()
                    cv2.waitKey(1)
                    break

            for box_cnt in range(self.crop_cnt):
                img = PlottingMixin().put_text(img=original_img, text=video_name, pos=txt_pos, font_size=font_size, font_thickness=2, font=self.font, text_color=(144, 238, 144), text_color_bg=(0, 0, 0))
                img = PlottingMixin().put_text(img=img, text=f"Draw crop box #{box_cnt+1} boundaries and press ESC", pos=(txt_pos[0], txt_pos[1] + (y_space * 1)), font_size=font_size, font_thickness=2, font=self.font, text_color=(144, 238, 144), text_color_bg=(0, 0, 0))
                roi_selector = ROISelector(path=img, title=f"VIDEO IMAGE")
                roi_selector.run()
                cv2.destroyAllWindows()
                if roi_selector.complete:
                    self.crop_df.loc[len(self.crop_df)] = [video_path, roi_selector.height, roi_selector.width, roi_selector.top_left[0], roi_selector.top_left[1]]
                else:
                    break

        self.__test_crop_locations()
        cv2.destroyAllWindows()
        cv2.waitKey(50)

        timer = SimbaTimer(start=True)
        crop_counter = 0
        print("Starting video cropping...")
        for video_path in self.crop_df["video"].unique():
            _, video_name, video_ext = get_fn_ext(filepath=video_path)
            video_crops = self.crop_df[self.crop_df["video"] == video_path].reset_index(drop=True)
            for cnt, (idx, row) in enumerate(video_crops.iterrows()):
                print(f"Creating {video_name} crop clip {cnt+1} (crop {crop_counter+1}/{len(self.crop_df)})...")
                crop_timer = SimbaTimer(start=True)
                height, width, top_left_x, top_left_y = row["height"], row["width"], row["top_left_x"], row["top_left_y"]
                save_path = os.path.join(self.output_folder, f"{video_name}_{cnt+1}.{self.file_type}")
                crop_video(video_path=video_path, save_path=save_path, size=(width, height), top_left=(top_left_x, top_left_y), gpu=self.gpu, verbose=False, quality=self.quality)
                crop_timer.stop_timer()
                print(f"Video {video_name} crop {cnt+1} complete (elapsed time: {crop_timer.elapsed_time_str})...")
                crop_counter += 1
        timer.stop_timer()
        stdout_success(msg=f"{str(len(self.crop_df))} new cropped videos created from {len(self.video_paths)} input videos. Cropped videos are saved in the {self.output_folder} directory", elapsed_time=timer.elapsed_time_str, source=self.__class__.__name__)


# cropper = MultiCropper(file_type='mp4', input_folder=r'C:\troubleshooting\mitra\test', output_folder=r'C:\troubleshooting\mitra\test\cropped', crop_cnt=2, gpu=True)
# cropper.run()


# cropper = MultiCropper(file_type='mp4', input_folder='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/videos', output_folder='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/edited', crop_cnt=2)
# cropper.run()



import functools
import multiprocessing
import os
import platform
from typing import Any, Dict, List, Optional, Union

import cv2
import numpy as np
from numba import jit, prange

from simba.mixins.config_reader import ConfigReader
from simba.mixins.geometry_mixin import GeometryMixin
from simba.mixins.plotting_mixin import PlottingMixin
from simba.utils.checks import (
    check_all_file_names_are_represented_in_video_log,
    check_file_exist_and_readable, check_filepaths_in_iterable_exist,
    check_float, check_if_keys_exist_in_dict, check_int, check_valid_lst)
from simba.utils.enums import Defaults, Formats, TagNames
from simba.utils.errors import NoSpecifiedOutputError
from simba.utils.printing import SimbaTimer, log_event, stdout_success
from simba.utils.read_write import (concatenate_videos_in_folder,
                                    find_core_cnt, get_fn_ext, read_df,
                                    remove_a_folder)

STYLE_PALETTE = 'palette'
STYLE_SHADING = 'shading'
STYLE_BIN_SIZE = 'bin_size'
STYLE_MAX_SCALE = 'max_scale'

STYLE_ATTR = [STYLE_PALETTE, STYLE_SHADING, STYLE_BIN_SIZE, STYLE_MAX_SCALE]

def _heatmap_location(data: np.array,
                      video_setting: bool,
                      frame_setting: bool,
                      video_temp_dir: str,
                      video_name: str,
                      frame_dir: str,
                      fps: int,
                      style_attr: dict,
                      aspect_ratio: float,
                      size: tuple,
                      make_location_heatmap_plot: PlottingMixin.make_location_heatmap_plot):

    group = int(data[0][0][1])
    if video_setting:
        fourcc = cv2.VideoWriter_fourcc(*Formats.MP4_CODEC.value)
        video_save_path = os.path.join(video_temp_dir, f"{group}.mp4")
        video_writer = cv2.VideoWriter(video_save_path, fourcc, fps, size)

    for i in range(data.shape[0]):
        frame_id = int(data[i, 0, 0])
        frm_data = data[i, :, 2:]

        img = PlottingMixin.make_location_heatmap_plot(frm_data=frm_data,
                                         max_scale=style_attr[STYLE_MAX_SCALE],
                                         palette=style_attr[STYLE_PALETTE],
                                         aspect_ratio=aspect_ratio,
                                         shading=style_attr[STYLE_SHADING],
                                         img_size=size,
                                         file_name=None)
        if video_setting:
            video_writer.write(img[:, :, :3])
        if frame_setting:
            file_path = os.path.join(frame_dir, "{}.png".format(frame_id))
            cv2.imwrite(file_path, img)
        print(f"Heatmap location/frame created: {frame_id + 1}, Video: {video_name}, Processing core: {group} ...")

    if video_setting:
        video_writer.release()

    return group


class HeatMapperLocationMultiprocess(ConfigReader, PlottingMixin):
    """
    Create heatmaps representing the location where animals spend time.

    ..note::
       `GitHub visualizations tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/tutorial.md#step-11-visualizations>`__.

    .. image:: _static/img/heatmap_location.gif
       :width: 500
       :align: center

    :param str config_path: path to SimBA project config file in Configparser format
    :param bool final_img_setting: If True, then create a single image representing the last frame of the input video
    :param bool video_setting: If True, then create a video of heatmaps.
    :param bool frame_setting: If True, then create individual heatmap frames.
    :param str clf_name: The name of the classified behavior.
    :param str bodypart: The name of the body-part used to infer the location of the classified behavior
    :param Dict style_attr: Dict containing settings for colormap, bin-size, max scale, and smooothing operations. For example: {'palette': 'jet', 'shading': 'gouraud', 'bin_size': 50, 'max_scale': 'auto'}.
    :param int core_cnt: The number of CPU cores to use. If -1, then all available cores.

    :example:
    >>> style_attr = {'palette': 'jet', 'shading': 'gouraud', 'bin_size': 100, 'max_scale': 'auto'}
    >>> heatmapper = HeatMapperLocationMultiprocess(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini', style_attr = style_attr, core_cnt=-1, final_img_setting=True, video_setting=True, frame_setting=False, bodypart='Nose_1', files_found=['/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/csv/machine_results/Together_1.csv'])
    >>> heatmapper.run()
    """

    def __init__(self,
                 config_path: Union[str, os.PathLike],
                 data_paths: List[Union[str, os.PathLike]],
                 bodypart: str,
                 style_attr: Dict[str, Any],
                 final_img_setting: Optional[bool] = True,
                 video_setting: Optional[bool] = False,
                 frame_setting: Optional[bool] = False,
                 core_cnt: Optional[int] = -1):

        if (not frame_setting) and (not video_setting) and (not final_img_setting):
            raise NoSpecifiedOutputError(msg="Please choose to select either heatmap videos, frames, and/or final image.", source=self.__class__.__name__)
        check_file_exist_and_readable(file_path=config_path)
        check_valid_lst(data=data_paths, valid_dtypes=(str,), min_len=1)
        check_if_keys_exist_in_dict(data=style_attr, key=STYLE_ATTR, name=f'{self.__class__.__name__} style_attr')
        check_int(name=f'{self.__class__.__name__} core_cnt', value=core_cnt, min_value=-1, max_value=find_core_cnt()[0])
        check_filepaths_in_iterable_exist(file_paths=data_paths, name=f'{self.__class__.__name__} data_paths')
        if core_cnt == -1: core_cnt = find_core_cnt()[0]
        self.core_cnt = core_cnt
        ConfigReader.__init__(self, config_path=config_path, create_logger=False)
        PlottingMixin.__init__(self)
        self.frame_setting, self.video_setting = frame_setting, video_setting
        self.final_img_setting, self.bp = final_img_setting, bodypart
        self.style_attr, self.data_paths = style_attr, data_paths
        if not os.path.exists(self.heatmap_location_dir):
            os.makedirs(self.heatmap_location_dir)
        self.bp_lst = [self.bp + "_x", self.bp + "_y"]
        if platform.system() == "Darwin":
            multiprocessing.set_start_method("spawn", force=True)
        print(f"Processing {len(self.data_paths)} video(s)...")

    @staticmethod
    @jit(nopython=True)
    def __insert_group_idx_column(data: np.array, group: int, last_frm_idx: int):
        results = np.full((data.shape[0], data.shape[1], data.shape[2] + 2), np.nan)
        group_col = np.full((data.shape[1], 1), group)
        for frm_idx in prange(data.shape[0]):
            h_stack = np.hstack((group_col, data[frm_idx]))
            frm_col = np.full((h_stack.shape[0], 1), frm_idx + last_frm_idx)
            results[frm_idx] = np.hstack((frm_col, h_stack))
        return results

    def run(self):
        check_all_file_names_are_represented_in_video_log(video_info_df=self.video_info_df, data_paths=self.data_paths)
        for file_cnt, file_path in enumerate(self.data_paths):
            video_timer = SimbaTimer(start=True)
            _, self.video_name, _ = get_fn_ext(file_path)
            self.video_info, self.px_per_mm, self.fps = self.read_video_info(video_name=self.video_name)
            self.width, self.height = int(self.video_info["Resolution_width"].values[0]), int(self.video_info["Resolution_height"].values[0])
            self.save_frame_folder_dir = os.path.join(self.heatmap_location_dir, self.video_name)
            self.video_folder = os.path.join(self.heatmap_location_dir, self.video_name)
            self.temp_folder = os.path.join(self.heatmap_location_dir, self.video_name, "temp")
            if self.frame_setting:
                if os.path.exists(self.save_frame_folder_dir):
                    remove_a_folder(folder_dir=self.save_frame_folder_dir)
                os.makedirs(self.save_frame_folder_dir)
            if self.video_setting:
                if os.path.exists(self.temp_folder):
                    remove_a_folder(folder_dir=self.temp_folder)
                    remove_a_folder(folder_dir=self.video_folder)
                os.makedirs(self.temp_folder)
                self.save_video_path = os.path.join(self.heatmap_location_dir, f"{self.video_name}.mp4")

            self.data_df = read_df(file_path=file_path, file_type=self.file_type, usecols=self.bp_lst)



            squares, aspect_ratio = GeometryMixin().bucket_img_into_grid_square(bucket_grid_size_mm=self.style_attr[STYLE_BIN_SIZE], img_size=(self.width, self.height), px_per_mm=self.px_per_mm)
            cum_sum_squares = GeometryMixin().cumsum_coord_geometries(data=self.data_df.values, fps=self.fps, geometries=squares)
            if self.style_attr[STYLE_MAX_SCALE] == "auto":
                self.style_attr[STYLE_MAX_SCALE]= np.round(np.max(np.max(cum_sum_squares[-1], axis=0)), 3)
                if self.style_attr[STYLE_MAX_SCALE] == 0: self.style_attr[STYLE_MAX_SCALE] = 1
            else:
                check_float(name=f'{self.__class__.__name__} style max scale', value=self.style_attr["max_scale"], min_value=10e-6)
                self.style_attr[STYLE_MAX_SCALE] = self.style_attr[STYLE_MAX_SCALE]

            if self.final_img_setting:
                self.make_location_heatmap_plot(frm_data=cum_sum_squares[-1, :, :],
                                                max_scale=self.style_attr[STYLE_MAX_SCALE],
                                                palette=self.style_attr[STYLE_PALETTE],
                                                aspect_ratio=aspect_ratio,
                                                file_name=os.path.join(self.heatmap_location_dir, f"{self.video_name}_final_frm.png"),
                                                shading=self.style_attr[STYLE_SHADING],
                                                img_size=(self.width, self.height))

            if self.video_setting or self.frame_setting:
                print("Creating heatmap location video frames...")
                frame_arrays = np.array_split(cum_sum_squares, self.core_cnt)
                last_frm_idx = 0
                for frm_group in range(len(frame_arrays)):
                    split_arr = frame_arrays[frm_group]
                    frame_arrays[frm_group] = self.__insert_group_idx_column(data=split_arr, group=frm_group, last_frm_idx=last_frm_idx)
                    last_frm_idx = np.max(frame_arrays[frm_group].reshape((frame_arrays[frm_group].shape[0], -1)))

                with multiprocessing.Pool(self.core_cnt, maxtasksperchild=self.maxtasksperchild) as pool:
                    constants = functools.partial(_heatmap_location,
                                                  video_setting=self.video_setting,
                                                  frame_setting=self.frame_setting,
                                                  style_attr=self.style_attr,
                                                  fps=self.fps,
                                                  video_temp_dir=self.temp_folder,
                                                  frame_dir=self.save_frame_folder_dir,
                                                  aspect_ratio=aspect_ratio,
                                                  size=(self.width, self.height),
                                                  video_name=self.video_name,
                                                  make_location_heatmap_plot=self.make_location_heatmap_plot)
                    for cnt, result in enumerate(pool.imap(constants,frame_arrays,chunksize=self.multiprocess_chunksize)):
                        print(f'Batch {result}/{self.core_cnt} complete... Video: {self.video_name} ({file_cnt+1}/{len(self.data_paths)})')
                    pool.terminate()
                    pool.join()

                if self.video_setting:
                    print(f"Joining {self.video_name} multiprocessed heatmap location video...")
                    concatenate_videos_in_folder(in_folder=self.temp_folder, save_path=self.save_video_path)

                video_timer.stop_timer()
                print(
                    "Heatmap video {} complete (elapsed time: {}s) ...".format(
                        self.video_name, video_timer.elapsed_time_str
                    )
                )

            self.timer.stop_timer()
            stdout_success(msg=f"Heatmap location videos visualizations for {len(self.data_paths)} videos created in {self.heatmap_location_dir} directory", elapsed_time=self.timer.elapsed_time_str, source=self.__class__.__name__)



# test = HeatMapperLocationMultiprocess(config_path='/Users/simon/Desktop/envs/simba/troubleshooting/platea/project_folder/project_config.ini',
#                                       style_attr = {'palette': 'magma', 'shading': 'gouraud', 'bin_size': 80, 'max_scale': 'auto'},
#                                       final_img_setting=True,
#                                       video_setting=True,
#                                       frame_setting=False,
#                                       bodypart='NOSE',
#                                       core_cnt=5,
#                                       data_paths=['/Users/simon/Desktop/envs/simba/troubleshooting/platea/project_folder/csv/outlier_corrected_movement_location/Video_1.csv'])
# test.run()



# test = HeatMapperLocationMultiprocess(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                                       style_attr = {'palette': 'jet', 'shading': 'gouraud', 'bin_size': 80, 'max_scale': 'auto'},
#                                       final_img_setting=True,
#                                       video_setting=False,
#                                       frame_setting=False,
#                                       bodypart='Nose_1',
#                                       core_cnt=5,
#                                       files_found=['/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/csv/outlier_corrected_movement_location/Together_1.csv'])
# test.run()

# test = HeatMapperLocationMultiprocess(config_path='/Users/simon/Desktop/envs/simba/troubleshooting/RAT_NOR/project_folder/project_config.ini',
#                                       style_attr = {'palette': 'magma', 'shading': 'gouraud', 'bin_size': 100, 'max_scale': 'auto'},
#                                       final_img_setting=True,
#                                       video_setting=True,
#                                       frame_setting=False,
#                                       bodypart='Nose',
#                                       core_cnt=5,
#                                       data_paths=['/Users/simon/Desktop/envs/simba/troubleshooting/RAT_NOR/project_folder/csv/machine_results/2022-06-20_NOB_DOT_4.csv'])
# test.run()

# img = np.zeros((img_size[0], img_size[1], 3)).astype(np.uint8)
# for i in polygons:
#     coords = np.array(i.exterior.coords).astype(np.int)
#     print(coords)
#     cv2.polylines(img, [coords], True, (0, 0, 50), 2)
# cv2.imshow('img', img)
# cv2.waitKey()



from itertools import product
from typing import Dict, List, Optional, Tuple

import numpy as np
import pandas as pd
import xgboost as xgb
from sklearn.model_selection import StratifiedKFold

from simba.model.regression.metrics import (mean_absolute_error,
                                            mean_absolute_percentage_error,
                                            mean_squared_error, r2_score,
                                            root_mean_squared_error)
from simba.utils.checks import (check_float, check_instance, check_int,
                                check_str, check_valid_array,
                                check_valid_dataframe, check_valid_lst)
from simba.utils.enums import Formats
from simba.utils.errors import DataHeaderError


def fit_xgb(x: pd.DataFrame,
            y: np.ndarray,
            mdl: xgb.XGBRegressor) -> xgb.XGBRegressor:
    """
    Fits an XGBoost regressor model to the given data.

    :param pd.DataFrame x: Input feature matrix where each row represents a sample and each column a feature. The data must have numeric types.
    :param np.ndarray y: Target values, must be a 1-dimensional array of numeric types with the same number  of rows as `x`.
    :param xgb.XGBRegressor mdl: Defined xgb.XGBRegressor. E.g., can be defined with :func:`simba.model.regression.model.xgb_define`,
    :return: Trained XGBoost regressor model.
    :rtype: xgb.XGBRegressor

    :example:
    >>> x = pd.DataFrame(np.random.randint(0, 500, (100, 20)))
    >>> y = np.random.randint(1, 6, (100,))
    >>> mdl = fit_xgb(x=x, y=y)
    """
    check_valid_dataframe(df=x, source=f'{fit_xgb.__name__} x', valid_dtypes=Formats.NUMERIC_DTYPES.value)
    check_valid_array(data=y, source=f'{fit_xgb.__name__} y', accepted_ndims=(1,), accepted_axis_0_shape=[x.shape[0]], accepted_dtypes=Formats.NUMERIC_DTYPES.value)
    check_instance(source=f'{fit_xgb.__name__} fit_xgb', instance=xgb_reg, accepted_types=(xgb.XGBRegressor,))
    return mdl.fit(X=x, y=y)

def transform_xgb(x: pd.DataFrame,
                  mdl: xgb.XGBRegressor) -> np.ndarray:

    """
    Transforms the input data using the provided XGBoost model by making predictions.

    :param pd.DataFrame x: Input feature matrix where each row represents a sample and each column a feature. The data must have numeric types.
    :param xgb.XGBRegressor mdl: Trained XGBoost model to use for making predictions.
    :return: Predictions rounded to 2 decimal places.
    :rtype: np.ndarray

    :example:
    >>> x, y = pd.DataFrame(np.random.randint(0, 500, (100, 20))), np.random.randint(1, 6, (100,))
    >>> mdl = fit_xgb(x=x, y=y)
    >>> new_x = pd.DataFrame(np.random.randint(0, 500, (100, 20)))
    >>> results = transform_xgb(x=new_x, mdl=mdl)

    :example:
    >>> x, y = pd.DataFrame(np.random.randint(0, 500, (100, 20))), np.random.randint(1, 6, (100,))
    >>> mdl = fit_xgb(x=x, y=y)
    >>> new_x = pd.DataFrame(np.random.randint(0, 500, (100, 20)))
    >>> results = transform_xgb(x=new_x, mdl=mdl)
    """

    check_instance(source=transform_xgb.__name__, instance=mdl, accepted_types=(xgb.XGBRegressor,))
    check_valid_dataframe(df=x, source=f'{transform_xgb.__name__} x', valid_dtypes=Formats.NUMERIC_DTYPES.value)
    expected_x_names = mdl.get_booster().feature_names
    new_x_names = [str(i) for i in list(x.columns)]
    missing_x_names = set([i for i in expected_x_names if i not in new_x_names])
    additional_x_names = set([i for i in new_x_names if i not in expected_x_names])
    if len(additional_x_names) > 0:
        raise DataHeaderError(msg=f'The new data has {len(additional_x_names)} features not expected by the model: {additional_x_names}', source=transform_xgb.__name__)
    if len(missing_x_names) > 0:
        raise DataHeaderError(msg=f'The new data are missing {len(missing_x_names)} features expected by the model: {missing_x_names}', source=transform_xgb.__name__)
    if expected_x_names != new_x_names:
        raise DataHeaderError(msg=f'The new data contains features in the wrong order from the expected features', source=transform_xgb.__name__)
    return np.round(mdl.predict(x), 2)


def evaluate_xgb(y_pred: np.ndarray,
                 y_true: np.ndarray,
                 metrics: List[str],
                 stratified: Optional[bool] = False) -> dict:
    """
    Evaluates the performance of a regression model (e.g., XGBoost) by calculating selected metrics. Optionally, the evaluation can be stratified by unique
    values in the true target variable (`y_true`), where performance is computed separately for each class/level.

    :param np.ndarray y_pred: Predicted values generated by the model, must have the same shape as `y_true`.
    :param np.ndarray y_true: True target values to compare the predictions against.
    :param List[str] metrics: List of metrics to compute.
    :param stratified: If True, computes the metric for each unique class/level in `y_true`. If False (default), computes the metric for the entire dataset.
    :return: A dictionary containing the computed metrics.
    :rtype: dict

    :example:
    >>> x = pd.DataFrame(np.random.randint(0, 500, (100, 20)))
    >>> y = np.random.randint(1, 6, (100,))
    >>> mdl = fit_xgb(x=x, y=y)
    >>> new_x = pd.DataFrame(np.random.randint(0, 500, (100, 20)))
    >>> y_pred = transform_xgb(x=new_x, mdl=mdl)
    >>> evaluate_xgb(y_pred=y_pred, y_true=y, metrics=['MAE', 'MAPE', 'RMSE', 'MSE'])
    """
    METRICS = {'MAPE': mean_absolute_percentage_error, 'MSE': mean_squared_error, 'MAE': mean_absolute_error, 'R2': r2_score, 'RMSE': root_mean_squared_error}
    check_valid_array(data=y_true, source=evaluate_xgb.__name__, accepted_ndims=(1,), accepted_dtypes=Formats.NUMERIC_DTYPES.value)
    check_valid_array(data=y_pred, source=evaluate_xgb.__name__, accepted_ndims=(1,), min_axis_0=y_true.shape[0],accepted_dtypes=Formats.NUMERIC_DTYPES.value)
    check_valid_lst(data=metrics, source=f'{evaluate_xgb.__name__} metrics', valid_values=list(METRICS.keys()))
    results = {}
    for metric in metrics:
        if not stratified:
            results[metric] = METRICS[metric](y_true=y_true, y_pred=y_pred)
        else:
            results[metric] = {}
            for unique_true in (np.unique(y_true)):
                sample_idx = np.argwhere(y_true == unique_true)
                sample_y_true, sample_y_pred = y_true[sample_idx].flatten(), y_pred[sample_idx].flatten()
                results[metric][unique_true] = METRICS[metric](y_true=sample_y_true, y_pred=sample_y_pred)

    return results

def xgb_define(objective: str = 'reg:squarederror',
               n_estimators: int = 100,
               max_depth: int = 6,
               verbosity: int = 1,
               learning_rate: float = 0.3,
               eta: float = 0.3,
               gamma: float = 0.0,
               tree_method: str = 'auto') -> xgb.XGBRegressor:

    """
    Defines an XGBoost regressor.

    :param str objective: The learning objective for the model.
    :param int n_estimators: Number of boosting rounds. Must be greater than or equal to 1. Default is 100.
    :param int max_depth: Maximum depth of a tree. Increasing this value makes the model more complex and more likely to overfit. Must be greater than or equal to 1. Default is 6.
    :param int verbosity: Verbosity of the training process (0-3).
    :param float learning_rate: Step size shrinkage used to prevent overfitting. Lower values make the model more robust but require more boosting rounds. Must be between 0.1 and 1.0. Default is 0.3.
    :param float eta: Learning rate alias. Must be between 0.0 and 1.0. Default is 0.3.
    :param float gamma: Minimum loss reduction required to make a further partition on a leaf node of the tree. Larger values prevent overfitting. Must be greater than or equal to 0.0. Default is 0.0.
    :param str tree_method: The tree construction algorithm used in XGBoost.
    :return: An initialized XGBoost Regressor with the specified configuration.
    :rtype: xgb.XGBRegressor
    """

    OBJECTIVES = ('reg:squarederror', 'reg:squaredlogerror', 'reg:logistic', 'reg:pseudohubererror')
    TREE_METHODS = ('auto', 'exact', 'approx', 'hist', 'gpu_hist')
    check_str(name=f'{fit_xgb.__name__} objective', value=objective, options=OBJECTIVES)
    check_str(name=f'{fit_xgb.__name__} tree_method', value=tree_method, options=TREE_METHODS)
    check_int(name=f'{fit_xgb.__name__} n_estimators', value=n_estimators, min_value=1)
    check_int(name=f'{fit_xgb.__name__} max_depth', value=max_depth, min_value=1)
    check_int(name=f'{fit_xgb.__name__} verbosity', value=verbosity, min_value=0, max_value=3)
    check_float(name=f'{fit_xgb.__name__} learning_rate', value=learning_rate, min_value=0.1, max_value=1.0)
    check_float(name=f'{fit_xgb.__name__} eta', value=eta, min_value=0.0, max_value=1.0)
    check_float(name=f'{fit_xgb.__name__} gamma', value=gamma, min_value=0.0)

    return xgb.XGBRegressor(objective=objective, max_depth=max_depth, n_estimators=n_estimators, verbosity=verbosity, learning_rate=learning_rate, eta=eta, gamma=gamma, tree_method=tree_method)


def xgb_grid_define(objective: Tuple[str] = ('reg:squarederror',),
                    n_estimators: Tuple[int] = (100,),
                    max_depth: Tuple[int] =(6,),
                    verbosity: Tuple[int] = (1,),
                    learning_rate: Tuple[float] = (0.3,),
                    eta: Tuple[float] = (0.3,),
                    gamma: Tuple[float] = (0.0,),
                    tree_method: Tuple[str] = ('auto',)) -> List[xgb.XGBRegressor]:

    grid = list(product(objective, n_estimators, max_depth, verbosity, learning_rate, eta, gamma, tree_method))
    mdls = []
    for i in grid:
        mdl = xgb_define(objective=i[0], n_estimators=i[1], max_depth=i[2], verbosity=i[3], learning_rate=i[4], eta=i[5], gamma=i[6], tree_method=i[7])
        mdls.append(mdl)
    return mdls


def xgb_grid_fit(x: pd.DataFrame,
                 y: np.ndarray,
                 mdls: List[xgb.XGBRegressor]) -> List[xgb.XGBRegressor]:

    check_valid_dataframe(df=x, source=f'{fit_xgb.__name__} x', valid_dtypes=Formats.NUMERIC_DTYPES.value)
    check_valid_array(data=y, source=f'{fit_xgb.__name__} y', accepted_ndims=(1,), accepted_axis_0_shape=[x.shape[0]], accepted_dtypes=Formats.NUMERIC_DTYPES.value)
    check_valid_lst(data=mdls, source=xgb_grid_fit.__name__, valid_dtypes=(xgb.XGBRegressor,))
    results = []
    for mdl in mdls:
        results.append(fit_xgb(x=x, y=y, mdl=mdl))
    return results






    # grid_df = pd.DataFrame(grid, columns=['objective', 'n_estimators', 'max_depth', 'learning_rate', 'eta', 'gamma'])
    # grid_df['verbosity'], grid_df['tree_method'] = verbosity, tree_method
    #
    # grid_df.apply(fit, axis=1, result_type='expand')








#xgb_grid_define(max_depth=(6, 3), gamma=(0, 0.3))

# x = pd.DataFrame(np.random.randint(0, 500, (100, 20)))
# y = np.random.randint(1, 6, (100,))
# mdl = fit_xgb(x=x, y=y)
# new_x = pd.DataFrame(np.random.randint(0, 500, (100, 20)))
# y_pred = transform_xgb(x=new_x, model=mdl)
# evaluate_xgb(y_pred=y_pred, y_true=y, metrics=['MAE', 'MAPE', 'RMSE', 'MSE'], stratified=True)
#


# x = pd.DataFrame(np.random.randint(0, 500, (100, 20)))
# y = np.random.randint(1, 6, (100,))
# #mdl = fit_xgb(x=x, y=y)
#
# kfold_fit_xgb(x=x, y=y)

    #for fold_cnt, (train_index, test_index) in enumerate(k_fold.split(x_data, y_data)):


from tkinter import *

from simba.mixins.pop_up_mixin import PopUpMixin
from simba.ui.get_tree_view import GetTreeView
from simba.ui.tkinter_functions import (CreateLabelFrameWithIcon, FileSelect,
                                        FolderSelect, SimbaButton)
from simba.utils.checks import (check_file_exist_and_readable,
                                check_if_dir_exists)
from simba.utils.enums import Options
from simba.utils.lookups import print_video_meta_data
from simba.utils.read_write import (find_all_videos_in_directory,
                                    get_video_meta_data)


class PrintVideoMetaDataPopUp(PopUpMixin):

    def __init__(self):

        PopUpMixin.__init__(self, title="PRINT VIDEO META DATA", icon='video')
        single_video_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="SINGLE VIDEO", icon_name='video')
        self.video_path = FileSelect(single_video_frm, "VIDEO PATH:", title="Select a video file", lblwidth=25, file_types=[("VIDEO", Options.ALL_VIDEO_FORMAT_STR_OPTIONS.value)])
        file_btn = SimbaButton(parent=single_video_frm, txt='RUN', txt_clr='navy', img='rocket', cmd=self.run, cmd_kwargs={'directory': False})

        multiple_video_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="MULTIPLE VIDEOS", icon_name='stack')
        self.directory_path = FolderSelect(multiple_video_frm, "VIDEO DIRECTORY PATH:",  title="Select folder with videos: ", lblwidth=25)
        dir_btn = SimbaButton(parent=multiple_video_frm, txt='RUN', txt_clr='navy', img='rocket', cmd=self.run, cmd_kwargs={'directory': True})

        single_video_frm.grid(row=0, column=0, sticky=NW)
        self.video_path.grid(row=0, column=0, sticky=NW)
        file_btn.grid(row=1, column=0, sticky=NW)

        multiple_video_frm.grid(row=1, column=0, sticky=NW)
        self.directory_path.grid(row=0, column=0, sticky=NW)
        dir_btn.grid(row=1, column=0, sticky=NW)
        self.main_frm.mainloop()

    def run(self, directory: bool):
        results = {}
        if directory:
            check_if_dir_exists(in_dir=self.directory_path.folder_path)
            videos = find_all_videos_in_directory(directory=self.directory_path.folder_path, as_dict=True)
            for video_name, video_path in videos.items():
                v = get_video_meta_data(video_path=video_path, fps_as_int=False)
                results[v['video_name']] = v
        else:
            check_file_exist_and_readable(file_path=self.video_path.file_path)
            v = get_video_meta_data(video_path=self.video_path.file_path, fps_as_int=False)
            results[v['video_name']] = v

        _ = GetTreeView(data=results, index_col_name='VIDEO', headers=('VALUE',), title='VIDEO META DATA')

#PrintVideoMetaDataPopUp()


import functools
import multiprocessing
import os
import shutil
from typing import Dict, List, Optional, Tuple, Union

import cv2
import numpy as np
import pandas as pd
from shapely.geometry import Polygon

from simba.data_processors.spontaneous_alternation_calculator import \
    SpontaneousAlternationCalculator
from simba.mixins.config_reader import ConfigReader
from simba.utils.checks import (check_file_exist_and_readable, check_float,
                                check_int, check_str, check_valid_lst)
from simba.utils.data import detect_bouts
from simba.utils.enums import Formats, Paths, TextOptions
from simba.utils.errors import AnimalNumberError, InvalidInputError
from simba.utils.printing import stdout_success
from simba.utils.read_write import (concatenate_videos_in_folder,
                                    find_core_cnt, find_video_of_file,
                                    get_video_meta_data, read_frm_of_video)

ALTERNATION_COLOR = (0, 255, 0)
ERROR_COLOR = (0, 0, 255)


def spontaneous_alternator_video_mp(
    frm_index: np.ndarray,
    video_path: Union[str, os.PathLike],
    temp_save_dir: Union[str, os.PathLike],
    event_txt: List[List[str]],
    alt_dict: Dict[str, List[int]],
    roi_geometries: Dict[str, Polygon],
    roi_geometry_clrs: Dict[str, Tuple[int]],
    animal_geometries: List[Polygon],
):

    core, frm_index = frm_index[0], frm_index[1:]
    video_meta_data = get_video_meta_data(video_path=video_path)
    cap = cv2.VideoCapture(video_path)
    start_frm, current_frm, end_frm = frm_index[0], frm_index[0], frm_index[-1]
    cap.set(1, start_frm)
    fourcc = cv2.VideoWriter_fourcc(*Formats.MP4_CODEC.value)
    save_path = os.path.join(temp_save_dir, f"{core}.mp4")
    writer = cv2.VideoWriter(
        save_path,
        fourcc,
        video_meta_data["fps"],
        (
            int(video_meta_data["width"] + (video_meta_data["width"] / 2)),
            video_meta_data["height"],
        ),
    )
    while current_frm < end_frm:
        while current_frm < end_frm:
            sequence_lst = event_txt[current_frm]
            border = np.zeros(
                (int(video_meta_data["height"]), int(video_meta_data["width"] / 2), 3),
                dtype=np.uint8,
            )
            ret, img = cap.read()
            for shape_cnt, (k, v) in enumerate(roi_geometries.items()):
                cv2.polylines(
                    img,
                    [np.array(v.exterior.coords).astype(np.int)],
                    True,
                    (roi_geometry_clrs[k]),
                    thickness=2,
                )
            cv2.polylines(
                img,
                [
                    np.array(animal_geometries[current_frm].exterior.coords).astype(
                        np.int
                    )
                ],
                True,
                (178, 102, 255),
                thickness=2,
            )
            if len(list(set(sequence_lst))) == len(list(roi_geometries.keys())) - 1:
                txt_clr = ALTERNATION_COLOR
            else:
                txt_clr = ERROR_COLOR
            error_1 = len(
                [x for x in alt_dict["same_arm_return_errors"] if x <= current_frm]
            )
            error_2 = len(
                [x for x in alt_dict["alt_arm_return_errors"] if x <= current_frm]
            )
            alternations = len(
                [x for x in alt_dict["alternations"] if x <= current_frm]
            )
            cv2.putText(
                border,
                "Sequence:" + ",".join(sequence_lst),
                (10, 50),
                TextOptions.FONT.value,
                1,
                txt_clr,
                2,
            )
            cv2.putText(
                border,
                f"Alternation #: {alternations}",
                (10, 80),
                TextOptions.FONT.value,
                1,
                txt_clr,
                2,
            )
            cv2.putText(
                border,
                f"Errors #: {error_1 + error_2}",
                (10, 110),
                TextOptions.FONT.value,
                1,
                txt_clr,
                2,
            )
            img = np.hstack((img, border))
            writer.write(img)
            print(f"Writing frame {current_frm} (Core: {core})...")
            current_frm += 1
    writer.release()


class SpontaneousAlternationsPlotter(ConfigReader):
    """
    Create plots representing delayed-alternation computations overlayed on video.

    .. image:: _static/img/SpontaneousAlternationsPlotter.gif
       :width: 700
       :align: center

    .. note::
       Uses ``simba.data_processors.spontaneous_alternation_calculator.SpontaneousAlternationCalculator`` to compute alternation statistics.

    :param Union[str, os.PathLike] config_path: Path to SimBA project config file.
    :param List[str] arm_names: List of ROI names representing the arms.
    :param str center_name: Name of the ROI representing the center of the maze
    :param Optional[int] animal_area: Value between 51 and 100, representing the percent of the animal body that has to be situated in a ROI for it to be considered an entry.
    :param Optional[float] threshold: Value between 0.0 and 1.0. Body-parts with detection probabilities below this value will be (if possible) filtered when constructing the animal geometry.
    :param Optional[int] buffer: Millimeters area for which the animal geometry should be increased in size. Useful if the animal geometry does not fully cover the animal.
    :param Optional[int] core_cnt: The number of CPU cores to use when creating the visualization. Defaults to -1 which represents all avaialbale cores.
    :param Optional[Union[str, os.PathLike]] data_path: Path to the file to be analyzed, e.g., CSV file in `project_folder/csv/outlier_corrected_movement_location`` directory.

    :example:
    >>> config_path = '/Users/simon/Desktop/envs/simba/troubleshooting/spontenous_alternation/project_folder/project_config.ini'
    >>> plotter = SpontaneousAlternationsPlotter(config_path=config_path, arm_names=['A', 'B', 'C'], center_name='Center', threshold=0.0, buffer=1, animal_area=60, data_path='/Users/simon/Desktop/envs/simba/troubleshooting/spontenous_alternation/project_folder/csv/outlier_corrected_movement_location/F1 HAB.csv')
    >>> plotter.run()
    """

    def __init__(
        self,
        config_path: Union[str, os.PathLike],
        arm_names: List[str],
        center_name: str,
        animal_area: Optional[int] = 80,
        threshold: Optional[float] = 0.0,
        buffer: Optional[int] = 2,
        core_cnt: Optional[int] = -1,
        verbose: Optional[bool] = False,
        data_path: Optional[Union[str, os.PathLike]] = None,
    ):

        ConfigReader.__init__(self, config_path=config_path)
        if self.animal_cnt != 1:
            raise AnimalNumberError(
                msg=f"Spontaneous alternation can only be calculated in 1 animal projects. Your project has {self.animal_cnt} animals.",
                source=self.__class__.__name__,
            )
        if len(self.body_parts_lst) < 3:
            raise InvalidInputError(
                msg=f"Spontaneous alternation can only be calculated in projects with 3 or more tracked body-parts. Found {len(self.body_parts_lst)}.",
                source=self.__class__.__name__,
            )
        check_valid_lst(
            data=arm_names,
            source=SpontaneousAlternationCalculator.__name__,
            valid_dtypes=(str,),
            min_len=2,
        )
        check_int(name="ANIMAL AREA", value=animal_area, min_value=1, max_value=100)
        check_float(name="THRESHOLD", value=threshold, min_value=0.0, max_value=1.0)
        check_int(name="BUFFER", value=buffer, min_value=1)
        check_file_exist_and_readable(file_path=data_path)
        check_int(
            name="CORE COUNT",
            value=core_cnt,
            min_value=-1,
            max_value=find_core_cnt()[0],
            raise_error=True,
        )
        check_str(name="CENTER NAME", value=center_name)
        if core_cnt == -1:
            core_cnt = find_core_cnt()[0]
        self.threshold, self.buffer, self.animal_area = threshold, buffer, animal_area
        self.verbose, self.arm_names, self.center_name = verbose, arm_names, center_name
        self.data_path, self.core_cnt = data_path, core_cnt

    def run(self):
        sa_computer = SpontaneousAlternationCalculator(
            config_path=self.config_path,
            arm_names=self.arm_names,
            center_name=self.center_name,
            animal_area=self.animal_area,
            threshold=self.threshold,
            verbose=False,
            buffer=self.buffer,
            data_path=self.data_path,
        )
        sa_computer.run()
        print("Running alternation visualization...")
        video_path = find_video_of_file(
            video_dir=self.video_dir, filename=sa_computer.video_name, raise_error=True
        )
        bout_df = (
            detect_bouts(
                data_df=sa_computer.roi_df,
                target_lst=self.arm_names + [self.center_name],
                fps=1,
            )[["Event", "Start_frame"]]
            .sort_values(["Start_frame"])
            .reset_index(drop=True)
        )
        shifted_ = pd.concat(
            [bout_df, bout_df.shift(-1).add_suffix("_shifted").reset_index(drop=True)],
            axis=1,
        )[["Event", "Event_shifted"]].values
        unique_counts = [len(list(set(list(x)))) for x in shifted_]
        drop_idx = np.argwhere(np.array(unique_counts) == 1) + 1
        bout_df = bout_df.drop(drop_idx.flatten(), axis=0).reset_index(drop=True)
        bout_df = bout_df[bout_df["Event"] != self.center_name]
        frm_index = np.arange(0, len(sa_computer.data_df))
        frm_index = np.array_split(frm_index, self.core_cnt)
        for cnt, i in enumerate(frm_index):
            frm_index[cnt] = np.insert(i, 0, cnt)
        event_txt = []
        for idx in range(len(sa_computer.data_df)):
            preceding_entries = list(
                bout_df["Event"][bout_df["Start_frame"] <= idx].tail(
                    len(self.arm_names)
                )
            )
            event_txt.append(preceding_entries)
        self.temp_folder = os.path.join(
            self.project_path,
            Paths.SPONTANEOUS_ALTERNATION_VIDEOS_DIR.value,
            sa_computer.video_name,
            "temp",
        )
        if os.path.isdir(self.temp_folder):
            shutil.rmtree(self.temp_folder)
        os.makedirs(self.temp_folder)
        alt_dict = {
            "same_arm_return_errors": [
                x
                for xs in sa_computer.video_results["same_arm_returns_dict"].values()
                for x in xs
            ],
            "alt_arm_return_errors": [
                x
                for xs in sa_computer.video_results[
                    "alternate_arm_returns_dict"
                ].values()
                for x in xs
            ],
            "alternations": [
                x
                for xs in sa_computer.video_results["alternations_dict"].values()
                for x in xs
            ],
        }
        save_path = os.path.join(
            self.project_path,
            Paths.SPONTANEOUS_ALTERNATION_VIDEOS_DIR.value,
            f"{sa_computer.video_name}.mp4",
        )
        with multiprocessing.Pool(
            self.core_cnt, maxtasksperchild=self.maxtasksperchild
        ) as pool:
            constants = functools.partial(
                spontaneous_alternator_video_mp,
                video_path=video_path,
                event_txt=event_txt,
                animal_geometries=sa_computer.animal_polygons,
                temp_save_dir=self.temp_folder,
                alt_dict=alt_dict,
                roi_geometries=sa_computer.roi_geos[sa_computer.video_name],
                roi_geometry_clrs=sa_computer.roi_clrs[sa_computer.video_name],
            )
            for cnt, result in enumerate(
                pool.imap(constants, frm_index, chunksize=self.multiprocess_chunksize)
            ):
                print(f"Section {cnt} complete...")
        pool.terminate()
        pool.join()
        print(f"Joining {sa_computer.video_name} multiprocessed video...")
        concatenate_videos_in_folder(in_folder=self.temp_folder, save_path=save_path)
        self.timer.stop_timer()
        stdout_success(
            f"Alternation video saved at {save_path}",
            elapsed_time=self.timer.elapsed_time_str,
        )


#
# x = SpontaneousAlternationsPlotter(
#     config_path='/Users/simon/Desktop/envs/simba/troubleshooting/spontenous_alternation/project_folder/project_config.ini',
#     arm_names=['A', 'B', 'C'],
#     center_name='Center',
#     threshold=0.0,
#     buffer=1,
#     animal_area=60,
#     data_path='/Users/simon/Desktop/envs/simba/troubleshooting/spontenous_alternation/project_folder/csv/outlier_corrected_movement_location/F1 HAB.csv')
# x.run()


#
